"""API Gateway Local Service"""
import base64
import io
import json
import logging
from datetime import datetime
from time import time
from typing import List, Optional

from flask import Flask, request
from werkzeug.datastructures import Headers
from werkzeug.routing import BaseConverter
from werkzeug.serving import WSGIRequestHandler

from samcli.commands.local.lib.exceptions import UnsupportedInlineCodeError
from samcli.lib.providers.provider import Cors
from samcli.lib.utils.stream_writer import StreamWriter
from samcli.local.events.api_event import (
    ApiGatewayLambdaEvent,
    ApiGatewayV2LambdaEvent,
    ContextHTTP,
    ContextIdentity,
    RequestContext,
    RequestContextV2,
)
from samcli.local.lambdafn.exceptions import FunctionNotFound
from samcli.local.services.base_local_service import BaseLocalService, LambdaOutputParser

from .path_converter import PathConverter
from .service_error_responses import ServiceErrorResponses

LOG = logging.getLogger(__name__)


class LambdaResponseParseException(Exception):
    """
    An exception raised when we fail to parse the response for Lambda
    """


class PayloadFormatVersionValidateException(Exception):
    """
    An exception raised when validation of payload format version fails
    """


class Route:
    API = "Api"
    HTTP = "HttpApi"
    ANY_HTTP_METHODS = ["GET", "DELETE", "PUT", "POST", "HEAD", "OPTIONS", "PATCH"]

    def __init__(
        self,
        function_name: Optional[str],
        path: str,
        methods: List[str],
        event_type: str = API,
        payload_format_version: Optional[str] = None,
        is_default_route: bool = False,
        operation_name=None,
        stack_path: str = "",
    ):
        """
        Creates an ApiGatewayRoute

        :param list(str) methods: http method
        :param function_name: Name of the Lambda function this API is connected to
        :param str path: Path off the base url
        :param str event_type: Type of the event. "Api" or "HttpApi"
        :param str payload_format_version: version of payload format
        :param bool is_default_route: determines if the default route or not
        :param string operation_name: Swagger operationId for the route
        :param str stack_path: path of the stack the route is located
        """
        self.methods = self.normalize_method(methods)
        self.function_name = function_name
        self.path = path
        self.event_type = event_type
        self.payload_format_version = payload_format_version
        self.is_default_route = is_default_route
        self.operation_name = operation_name
        self.stack_path = stack_path

    def __eq__(self, other):
        return (
            isinstance(other, Route)
            and sorted(self.methods) == sorted(other.methods)
            and self.function_name == other.function_name
            and self.path == other.path
            and self.operation_name == other.operation_name
            and self.stack_path == other.stack_path
        )

    def __hash__(self):
        route_hash = hash(f"{self.stack_path}-{self.function_name}-{self.path}")
        for method in sorted(self.methods):
            route_hash *= hash(method)
        return route_hash

    def normalize_method(self, methods):
        """
        Normalizes Http Methods. Api Gateway allows a Http Methods of ANY. This is a special verb to denote all
        supported Http Methods on Api Gateway.

        :param list methods: Http methods
        :return list: Either the input http_method or one of the _ANY_HTTP_METHODS (normalized Http Methods)
        """
        methods = [method.upper() for method in methods]
        if "ANY" in methods:
            return self.ANY_HTTP_METHODS
        return methods


class CatchAllPathConverter(BaseConverter):
    regex = ".+"
    weight = 300
    part_isolating = False

    def to_python(self, value):
        return value

    def to_url(self, value):
        return value


class LocalApigwService(BaseLocalService):
    _DEFAULT_PORT = 3000
    _DEFAULT_HOST = "127.0.0.1"

    def __init__(self, api, lambda_runner, static_dir=None, port=None, host=None, stderr=None):
        """
        Creates an ApiGatewayService

        Parameters
        ----------
        api : Api
           an Api object that contains the list of routes and properties
        lambda_runner : samcli.commands.local.lib.local_lambda.LocalLambdaRunner
            The Lambda runner class capable of invoking the function
        static_dir : str
            Directory from which to serve static files
        port : int
            Optional. port for the service to start listening on
            Defaults to 3000
        host : str
            Optional. host to start the service on
            Defaults to '127.0.0.1
        stderr : samcli.lib.utils.stream_writer.StreamWriter
            Optional stream writer where the stderr from Docker container should be written to
        """
        super().__init__(lambda_runner.is_debugging(), port=port, host=host)
        self.api = api
        self.lambda_runner = lambda_runner
        self.static_dir = static_dir
        self._dict_of_routes = {}
        self.stderr = stderr

    def create(self):
        """
        Creates a Flask Application that can be started.
        """
        # Setting sam local start-api to respond using HTTP/1.1 instead of the default HTTP/1.0
        WSGIRequestHandler.protocol_version = "HTTP/1.1"

        self._app = Flask(
            __name__,
            static_url_path="",  # Mount static files at root '/'
            static_folder=self.static_dir,  # Serve static files from this directory
        )

        # add converter to support catch-all route
        self._app.url_map.converters["path"] = CatchAllPathConverter

        # Prevent the dev server from emitting headers that will make the browser cache response by default
        self._app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0

        # This will normalize all endpoints and strip any trailing '/'
        self._app.url_map.strict_slashes = False
        default_route = None

        for api_gateway_route in self.api.routes:
            if api_gateway_route.path == "$default":
                default_route = api_gateway_route
                continue
            path = PathConverter.convert_path_to_flask(api_gateway_route.path)
            for route_key in self._generate_route_keys(api_gateway_route.methods, path):
                self._dict_of_routes[route_key] = api_gateway_route
            self._app.add_url_rule(
                path,
                endpoint=path,
                view_func=self._request_handler,
                methods=api_gateway_route.methods,
                provide_automatic_options=False,
            )

        if default_route:
            LOG.debug("add catch-all route")
            all_methods = Route.ANY_HTTP_METHODS
            try:
                rules_iter = self._app.url_map.iter_rules("/")
                while True:
                    rule = next(rules_iter)
                    all_methods = [method for method in all_methods if method not in rule.methods]
            except (KeyError, StopIteration):
                pass

            self._add_catch_all_path(all_methods, "/", default_route)
            self._add_catch_all_path(Route.ANY_HTTP_METHODS, "/<path:any_path>", default_route)

        self._construct_error_handling()

    def _add_catch_all_path(self, methods, path, route):
        """
        Add the catch all route to the _app and the dictionary of routes.

        :param list(str) methods: List of HTTP Methods
        :param str path: Path off the base url
        :param Route route: contains the default route configurations
        """

        self._app.add_url_rule(
            path,
            endpoint=path,
            view_func=self._request_handler,
            methods=methods,
            provide_automatic_options=False,
        )
        for route_key in self._generate_route_keys(methods, path):
            self._dict_of_routes[route_key] = Route(
                function_name=route.function_name,
                path=path,
                methods=methods,
                event_type=Route.HTTP,
                payload_format_version=route.payload_format_version,
                is_default_route=True,
                stack_path=route.stack_path,
            )

    def _generate_route_keys(self, methods, path):
        """
        Generates the key to the _dict_of_routes based on the list of methods
        and path supplied

        Parameters
        ----------
        methods : List[str]
            List of HTTP Methods
        path : str
            Path off the base url

        Yields
        ------
        route_key : str
            the route key in the form of "Path:Method"
        """
        for method in methods:
            yield self._route_key(method, path)

    @staticmethod
    def _v2_route_key(method, path, is_default_route):
        if is_default_route:
            return "$default"
        return "{} {}".format(method, path)

    @staticmethod
    def _route_key(method, path):
        return "{}:{}".format(path, method)

    def _construct_error_handling(self):
        """
        Updates the Flask app with Error Handlers for different Error Codes
        """
        # Both path and method not present
        self._app.register_error_handler(404, ServiceErrorResponses.route_not_found)
        # Path is present, but method not allowed
        self._app.register_error_handler(405, ServiceErrorResponses.route_not_found)
        # Something went wrong
        self._app.register_error_handler(500, ServiceErrorResponses.lambda_failure_response)

    def _request_handler(self, **kwargs):
        """
        We handle all requests to the host:port. The general flow of handling a request is as follows

        * Fetch request from the Flask Global state. This is where Flask places the request and is per thread so
          multiple requests are still handled correctly
        * Find the Lambda function to invoke by doing a look up based on the request.endpoint and method
        * If we don't find the function, we will throw a 502 (just like the 404 and 405 responses we get
          from Flask.
        * Since we found a Lambda function to invoke, we construct the Lambda Event from the request
        * Then Invoke the Lambda function (docker container)
        * We then transform the response or errors we get from the Invoke and return the data back to
          the caller

        Parameters
        ----------
        kwargs dict
            Keyword Args that are passed to the function from Flask. This happens when we have path parameters

        Returns
        -------
        Response object
        """

        route = self._get_current_route(request)
        cors_headers = Cors.cors_to_headers(self.api.cors)

        # payloadFormatVersion can only support 2 values: "1.0" and "2.0"
        # so we want to do strict validation to make sure it has proper value if provided
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html
        if route.payload_format_version not in [None, "1.0", "2.0"]:
            raise PayloadFormatVersionValidateException(
                f'{route.payload_format_version} is not a valid value. PayloadFormatVersion must be "1.0" or "2.0"'
            )

        method, endpoint = self.get_request_methods_endpoints(request)
        if method == "OPTIONS" and self.api.cors:
            headers = Headers(cors_headers)
            return self.service_response("", headers, 200)

        try:
            # TODO: Rewrite the logic below to use version 2.0 when an invalid value is provided
            # the Lambda Event 2.0 is only used for the HTTP API gateway with defined payload format version equal 2.0
            # or none, as the default value to be used is 2.0
            # https://docs.aws.amazon.com/apigatewayv2/latest/api-reference/apis-apiid-integrations.html#apis-apiid-integrations-prop-createintegrationinput-payloadformatversion
            if route.event_type == Route.HTTP and route.payload_format_version in [None, "2.0"]:
                apigw_endpoint = PathConverter.convert_path_to_api_gateway(endpoint)
                route_key = self._v2_route_key(method, apigw_endpoint, route.is_default_route)
                event = self._construct_v_2_0_event_http(
                    request,
                    self.port,
                    self.api.binary_media_types,
                    self.api.stage_name,
                    self.api.stage_variables,
                    route_key,
                )
            elif route.event_type == Route.API:
                # The OperationName is only sent to the Lambda Function from API Gateway V1(Rest API).
                event = self._construct_v_1_0_event(
                    request,
                    self.port,
                    self.api.binary_media_types,
                    self.api.stage_name,
                    self.api.stage_variables,
                    route.operation_name,
                )
            else:
                # For Http Apis with payload version 1.0, API Gateway never sends the OperationName.
                event = self._construct_v_1_0_event(
                    request,
                    self.port,
                    self.api.binary_media_types,
                    self.api.stage_name,
                    self.api.stage_variables,
                    None,
                )
        except UnicodeDecodeError as error:
            LOG.error("UnicodeDecodeError while processing HTTP request: %s", error)
            return ServiceErrorResponses.lambda_failure_response()

        stdout_stream = io.BytesIO()
        stdout_stream_writer = StreamWriter(stdout_stream, auto_flush=True)

        try:
            self.lambda_runner.invoke(route.function_name, event, stdout=stdout_stream_writer, stderr=self.stderr)
        except FunctionNotFound:
            return ServiceErrorResponses.lambda_not_found_response()
        except UnsupportedInlineCodeError:
            return ServiceErrorResponses.not_implemented_locally(
                "Inline code is not supported for sam local commands. Please write your code in a separate file."
            )

        lambda_response, _ = LambdaOutputParser.get_lambda_output(stdout_stream)

        try:
            if route.event_type == Route.HTTP and (
                not route.payload_format_version or route.payload_format_version == "2.0"
            ):
                (status_code, headers, body) = self._parse_v2_payload_format_lambda_output(
                    lambda_response, self.api.binary_media_types, request
                )
            else:
                (status_code, headers, body) = self._parse_v1_payload_format_lambda_output(
                    lambda_response, self.api.binary_media_types, request, route.event_type
                )
        except LambdaResponseParseException as ex:
            LOG.error("Invalid lambda response received: %s", ex)
            return ServiceErrorResponses.lambda_failure_response()

        return self.service_response(body, headers, status_code)

    def _get_current_route(self, flask_request):
        """
        Get the route (Route) based on the current request

        :param request flask_request: Flask Request
        :return: Route matching the endpoint and method of the request
        """
        method, endpoint = self.get_request_methods_endpoints(flask_request)

        route_key = self._route_key(method, endpoint)
        route = self._dict_of_routes.get(route_key, None)

        if not route:
            LOG.debug(
                "Lambda function for the route not found. This should not happen because Flask is "
                "already configured to serve all path/methods given to the service. "
                "Path=%s Method=%s RouteKey=%s",
                endpoint,
                method,
                route_key,
            )
            raise KeyError("Lambda function for the route not found")

        return route

    @staticmethod
    def get_request_methods_endpoints(flask_request):
        """
        Separated out for testing requests in request handler
        :param request flask_request: Flask Request
        :return: the request's endpoint and method
        """
        endpoint = flask_request.endpoint
        method = flask_request.method
        return method, endpoint

    # Consider moving this out to its own class. Logic is started to get dense and looks messy @jfuss
    @staticmethod
    def _parse_v1_payload_format_lambda_output(lambda_output: str, binary_types, flask_request, event_type):
        """
        Parses the output from the Lambda Container

        :param str lambda_output: Output from Lambda Invoke
        :param binary_types: list of binary types
        :param flask_request: flash request object
        :param event_type: determines the route event type
        :return: Tuple(int, dict, str, bool)
        """
        # pylint: disable-msg=too-many-statements
        try:
            json_output = json.loads(lambda_output)
        except ValueError as ex:
            raise LambdaResponseParseException("Lambda response must be valid json") from ex

        if not isinstance(json_output, dict):
            raise LambdaResponseParseException(f"Lambda returned {type(json_output)} instead of dict")

        if event_type == Route.HTTP and json_output.get("statusCode") is None:
            raise LambdaResponseParseException(f"Invalid API Gateway Response Key: statusCode is not in {json_output}")

        status_code = json_output.get("statusCode") or 200
        headers = LocalApigwService._merge_response_headers(
            json_output.get("headers") or {}, json_output.get("multiValueHeaders") or {}
        )

        body = json_output.get("body")
        if body is None:
            LOG.warning("Lambda returned empty body!")

        is_base_64_encoded = LocalApigwService.get_base_64_encoded(event_type, json_output)

        try:
            status_code = int(status_code)
            if status_code <= 0:
                raise ValueError
        except ValueError as ex:
            raise LambdaResponseParseException("statusCode must be a positive int") from ex

        try:
            if body:
                body = str(body)
        except ValueError as ex:
            raise LambdaResponseParseException(
                f"Non null response bodies should be able to convert to string: {body}"
            ) from ex

        invalid_keys = LocalApigwService._invalid_apig_response_keys(json_output, event_type)
        # HTTP API Gateway just skip the non allowed lambda response fields, but Rest API gateway fail on
        # the non allowed fields
        if event_type == Route.API and invalid_keys:
            raise LambdaResponseParseException(f"Invalid API Gateway Response Keys: {invalid_keys} in {json_output}")

        # If the customer doesn't define Content-Type default to application/json
        if "Content-Type" not in headers:
            LOG.info("No Content-Type given. Defaulting to 'application/json'.")
            headers["Content-Type"] = "application/json"

        try:
            # HTTP API Gateway always decode the lambda response only if isBase64Encoded field in response is True
            # regardless the response content-type
            # Rest API Gateway depends on the response content-type and the API configured BinaryMediaTypes to decide
            # if it will decode the response or not
            if (event_type == Route.HTTP and is_base_64_encoded) or (
                event_type == Route.API
                and LocalApigwService._should_base64_decode_body(
                    binary_types, flask_request, headers, is_base_64_encoded
                )
            ):
                body = base64.b64decode(body)
        except ValueError as ex:
            LambdaResponseParseException(str(ex))

        return status_code, headers, body

    @staticmethod
    def get_base_64_encoded(event_type, json_output):
        # The following behaviour is undocumented behaviour, and based on some trials
        # Http API gateway checks lambda response for isBase64Encoded field, and ignore base64Encoded
        # Rest API gateway checks first the field base64Encoded field, if not exist, it checks isBase64Encoded field

        if event_type == Route.API and json_output.get("base64Encoded") is not None:
            is_base_64_encoded = json_output.get("base64Encoded")
            field_name = "base64Encoded"
        elif json_output.get("isBase64Encoded") is not None:
            is_base_64_encoded = json_output.get("isBase64Encoded")
            field_name = "isBase64Encoded"
        else:
            is_base_64_encoded = False
            field_name = "isBase64Encoded"

        if isinstance(is_base_64_encoded, str) and is_base_64_encoded in ["true", "True", "false", "False"]:
            is_base_64_encoded = is_base_64_encoded in ["true", "True"]
        elif not isinstance(is_base_64_encoded, bool):
            raise LambdaResponseParseException(
                f"Invalid API Gateway Response Key: {is_base_64_encoded} is not a valid" f"{field_name}"
            )

        return is_base_64_encoded

    @staticmethod
    def _parse_v2_payload_format_lambda_output(lambda_output: str, binary_types, flask_request):
        """
        Parses the output from the Lambda Container. V2 Payload Format means that the event_type is only HTTP

        :param str lambda_output: Output from Lambda Invoke
        :param binary_types: list of binary types
        :param flask_request: flash request object
        :return: Tuple(int, dict, str, bool)
        """
        # pylint: disable-msg=too-many-statements
        # pylint: disable=too-many-branches
        try:
            json_output = json.loads(lambda_output)
        except ValueError as ex:
            raise LambdaResponseParseException("Lambda response must be valid json") from ex

        # lambda can return any valid json response in payload format version 2.0.
        # response can be a simple type like string, or integer
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html#http-api-develop-integrations-lambda.response
        if isinstance(json_output, dict):
            body = json_output.get("body") if "statusCode" in json_output else json.dumps(json_output)
        else:
            body = json_output
            json_output = {}

        if body is None:
            LOG.warning("Lambda returned empty body!")

        status_code = json_output.get("statusCode") or 200
        headers = Headers(json_output.get("headers") or {})

        # cookies is a new field in payload format version 2.0 (a list)
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html
        # we need to move cookies to Set-Cookie headers.
        # each cookie becomes a set-cookie header
        # MDN link: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie
        cookies = json_output.get("cookies")

        # cookies needs to be a list, otherwise the format is wrong and we can skip it
        if isinstance(cookies, list):
            for cookie in cookies:
                headers.add("Set-Cookie", cookie)

        is_base_64_encoded = json_output.get("isBase64Encoded") or False

        try:
            status_code = int(status_code)
            if status_code <= 0:
                raise ValueError
        except ValueError as ex:
            raise LambdaResponseParseException("statusCode must be a positive int") from ex

        try:
            if body:
                body = str(body)
        except ValueError as ex:
            raise LambdaResponseParseException(
                f"Non null response bodies should be able to convert to string: {body}"
            ) from ex

        # If the customer doesn't define Content-Type default to application/json
        if "Content-Type" not in headers:
            LOG.info("No Content-Type given. Defaulting to 'application/json'.")
            headers["Content-Type"] = "application/json"

        try:
            # HTTP API Gateway always decode the lambda response only if isBase64Encoded field in response is True
            # regardless the response content-type
            if is_base_64_encoded:
                # Note(xinhol): here in this method we change the type of the variable body multiple times
                # and confused mypy, we might want to avoid this and use multiple variables here.
                body = base64.b64decode(body)  # type: ignore
        except ValueError as ex:
            LambdaResponseParseException(str(ex))

        return status_code, headers, body

    @staticmethod
    def _invalid_apig_response_keys(output, event_type):
        allowable = {"statusCode", "body", "headers", "multiValueHeaders", "isBase64Encoded", "cookies"}
        if event_type == Route.API:
            allowable.add("base64Encoded")
        invalid_keys = output.keys() - allowable
        return invalid_keys

    @staticmethod
    def _should_base64_decode_body(binary_types, flask_request, lamba_response_headers, is_base_64_encoded):
        """
        Whether or not the body should be decoded from Base64 to Binary

        Parameters
        ----------
        binary_types list(basestring)
            Corresponds to self.binary_types (aka. what is parsed from SAM Template
        flask_request flask.request
            Flask request
        lamba_response_headers werkzeug.datastructures.Headers
            Headers Lambda returns
        is_base_64_encoded bool
            True if the body is Base64 encoded

        Returns
        -------
        True if the body from the request should be converted to binary, otherwise false

        """
        best_match_mimetype = flask_request.accept_mimetypes.best_match(lamba_response_headers.get_all("Content-Type"))
        is_best_match_in_binary_types = best_match_mimetype in binary_types or "*/*" in binary_types

        return best_match_mimetype and is_best_match_in_binary_types and is_base_64_encoded

    @staticmethod
    def _merge_response_headers(headers, multi_headers):
        """
        Merge multiValueHeaders headers with headers

        * If you specify values for both headers and multiValueHeaders, API Gateway merges them into a single list.
        * If the same key-value pair is specified in both, the value will only appear once.

        Parameters
        ----------
        headers dict
            Headers map from the lambda_response_headers
        multi_headers dict
            multiValueHeaders map from the lambda_response_headers

        Returns
        -------
        Merged list in accordance to the AWS documentation within a Flask Headers object

        """

        processed_headers = Headers(multi_headers)

        for header in headers:
            # Prevent duplication of values when the key-value pair exists in both
            # headers and multi_headers, but preserve order from multi_headers
            if header in multi_headers and headers[header] in multi_headers[header]:
                continue

            processed_headers.add(header, headers[header])

        return processed_headers

    @staticmethod
    def _construct_v_1_0_event(
        flask_request, port, binary_types, stage_name=None, stage_variables=None, operation_name=None
    ):
        """
        Helper method that constructs the Event to be passed to Lambda

        :param request flask_request: Flask Request
        :param port: the port number
        :param binary_types: list of binary types
        :param stage_name: Optional, the stage name string
        :param stage_variables: Optional, API Gateway Stage Variables
        :return: String representing the event
        """
        # pylint: disable-msg=too-many-locals

        identity = ContextIdentity(source_ip=flask_request.remote_addr)

        endpoint = PathConverter.convert_path_to_api_gateway(flask_request.endpoint)
        method = flask_request.method
        protocol = flask_request.environ.get("SERVER_PROTOCOL", "HTTP/1.1")
        host = flask_request.host

        request_data = flask_request.get_data()

        request_mimetype = flask_request.mimetype

        is_base_64 = LocalApigwService._should_base64_encode(binary_types, request_mimetype)

        if is_base_64:
            LOG.debug("Incoming Request seems to be binary. Base64 encoding the request data before sending to Lambda.")
            request_data = base64.b64encode(request_data)

        if request_data:
            # Flask does not parse/decode the request data. We should do it ourselves
            # Note(xinhol): here we change request_data's type from bytes to str and confused mypy
            # We might want to consider to use a new variable here.
            request_data = request_data.decode("utf-8")

        query_string_dict, multi_value_query_string_dict = LocalApigwService._query_string_params(flask_request)

        context = RequestContext(
            resource_path=endpoint,
            http_method=method,
            stage=stage_name,
            identity=identity,
            path=endpoint,
            protocol=protocol,
            domain_name=host,
            operation_name=operation_name,
        )

        headers_dict, multi_value_headers_dict = LocalApigwService._event_headers(flask_request, port)

        event = ApiGatewayLambdaEvent(
            http_method=method,
            body=request_data,
            resource=endpoint,
            request_context=context,
            query_string_params=query_string_dict,
            multi_value_query_string_params=multi_value_query_string_dict,
            headers=headers_dict,
            multi_value_headers=multi_value_headers_dict,
            path_parameters=flask_request.view_args,
            path=flask_request.path,
            is_base_64_encoded=is_base_64,
            stage_variables=stage_variables,
        )

        event_str = json.dumps(event.to_dict(), sort_keys=True)
        LOG.debug("Constructed String representation of Event to invoke Lambda. Event: %s", event_str)
        return event_str

    @staticmethod
    def _construct_v_2_0_event_http(
        flask_request,
        port,
        binary_types,
        stage_name=None,
        stage_variables=None,
        route_key=None,
        request_time_epoch=int(time()),
        request_time=datetime.utcnow().strftime("%d/%b/%Y:%H:%M:%S +0000"),
    ):
        """
        Helper method that constructs the Event 2.0 to be passed to Lambda

        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html

        :param request flask_request: Flask Request
        :param port: the port number
        :param binary_types: list of binary types
        :param stage_name: Optional, the stage name string
        :param stage_variables: Optional, API Gateway Stage Variables
        :param route_key: Optional, the route key for the route
        :return: String representing the event
        """
        # pylint: disable-msg=too-many-locals
        method = flask_request.method

        request_data = flask_request.get_data()

        request_mimetype = flask_request.mimetype

        is_base_64 = LocalApigwService._should_base64_encode(binary_types, request_mimetype)

        if is_base_64:
            LOG.debug("Incoming Request seems to be binary. Base64 encoding the request data before sending to Lambda.")
            request_data = base64.b64encode(request_data)

        if request_data is not None:
            # Flask does not parse/decode the request data. We should do it ourselves
            request_data = request_data.decode("utf-8")

        query_string_dict = LocalApigwService._query_string_params_v_2_0(flask_request)

        cookies = LocalApigwService._event_http_cookies(flask_request)
        headers = LocalApigwService._event_http_headers(flask_request, port)
        context_http = ContextHTTP(method=method, path=flask_request.path, source_ip=flask_request.remote_addr)
        context = RequestContextV2(
            http=context_http,
            route_key=route_key,
            stage=stage_name,
            request_time_epoch=request_time_epoch,
            request_time=request_time,
        )

        event = ApiGatewayV2LambdaEvent(
            route_key=route_key,
            raw_path=flask_request.path,
            raw_query_string=flask_request.query_string.decode("utf-8"),
            cookies=cookies,
            headers=headers,
            query_string_params=query_string_dict,
            request_context=context,
            body=request_data,
            path_parameters=flask_request.view_args,
            is_base_64_encoded=is_base_64,
            stage_variables=stage_variables,
        )

        event_str = json.dumps(event.to_dict())
        LOG.debug("Constructed String representation of Event Version 2.0 to invoke Lambda. Event: %s", event_str)
        return event_str

    @staticmethod
    def _query_string_params(flask_request):
        """
        Constructs an APIGW equivalent query string dictionary

        Parameters
        ----------
        flask_request request
            Request from Flask

        Returns dict (str: str), dict (str: list of str)
        -------
            Empty dict if no query params where in the request otherwise returns a dictionary of key to value

        """
        query_string_dict = {}
        multi_value_query_string_dict = {}

        # Flask returns an ImmutableMultiDict so convert to a dictionary that becomes
        # a dict(str: list) then iterate over
        for query_string_key, query_string_list in flask_request.args.lists():
            query_string_value_length = len(query_string_list)

            # if the list is empty, default to empty string
            if not query_string_value_length:
                query_string_dict[query_string_key] = ""
                multi_value_query_string_dict[query_string_key] = [""]
            else:
                query_string_dict[query_string_key] = query_string_list[-1]
                multi_value_query_string_dict[query_string_key] = query_string_list

        return query_string_dict, multi_value_query_string_dict

    @staticmethod
    def _query_string_params_v_2_0(flask_request):
        """
        Constructs an APIGW equivalent query string dictionary using the 2.0 format
        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html#2.0

        Parameters
        ----------
        flask_request request
            Request from Flask

        Returns dict (str: str)
        -------
            Empty dict if no query params where in the request otherwise returns a dictionary of key to value

        """
        query_string_dict = {}

        # Flask returns an ImmutableMultiDict so convert to a dictionary that becomes
        # a dict(str: list) then iterate over
        query_string_dict = {
            query_string_key: ",".join(query_string_list)
            for query_string_key, query_string_list in flask_request.args.lists()
        }

        return query_string_dict

    @staticmethod
    def _event_headers(flask_request, port):
        """
        Constructs an APIGW equivalent headers dictionary

        Parameters
        ----------
        flask_request request
            Request from Flask
        int port
            Forwarded Port
        cors_headers dict
            Dict of the Cors properties

        Returns dict (str: str), dict (str: list of str)
        -------
            Returns a dictionary of key to list of strings

        """
        headers_dict = {}
        multi_value_headers_dict = {}

        # Multi-value request headers is not really supported by Flask.
        # See https://github.com/pallets/flask/issues/850
        for header_key in flask_request.headers.keys():
            headers_dict[header_key] = flask_request.headers.get(header_key)
            multi_value_headers_dict[header_key] = flask_request.headers.getlist(header_key)

        headers_dict["X-Forwarded-Proto"] = flask_request.scheme
        multi_value_headers_dict["X-Forwarded-Proto"] = [flask_request.scheme]

        headers_dict["X-Forwarded-Port"] = str(port)
        multi_value_headers_dict["X-Forwarded-Port"] = [str(port)]
        return headers_dict, multi_value_headers_dict

    @staticmethod
    def _event_http_cookies(flask_request):
        """
        All cookie headers in the request are combined with commas.

        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html

        Parameters
        ----------
        flask_request request
            Request from Flask

        Returns list
        -------
            Returns a list of cookies

        """
        cookies = []
        for cookie_key in flask_request.cookies.keys():
            cookies.append("{}={}".format(cookie_key, flask_request.cookies.get(cookie_key)))
        return cookies

    @staticmethod
    def _event_http_headers(flask_request, port):
        """
        Duplicate headers are combined with commas.

        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html

        Parameters
        ----------
        flask_request request
            Request from Flask

        Returns list
        -------
            Returns a list of cookies

        """
        headers = {}
        # Multi-value request headers is not really supported by Flask.
        # See https://github.com/pallets/flask/issues/850
        for header_key in flask_request.headers.keys():
            headers[header_key] = flask_request.headers.get(header_key)

        headers["X-Forwarded-Proto"] = flask_request.scheme
        headers["X-Forwarded-Port"] = str(port)
        return headers

    @staticmethod
    def _should_base64_encode(binary_types, request_mimetype):
        """
        Whether or not to encode the data from the request to Base64

        Parameters
        ----------
        binary_types list(basestring)
            Corresponds to self.binary_types (aka. what is parsed from SAM Template
        request_mimetype str
            Mimetype for the request

        Returns
        -------
            True if the data should be encoded to Base64 otherwise False

        """
        return request_mimetype in binary_types or "*/*" in binary_types



"""
Base class for SAM Template providers
"""

import logging
from typing import Any, Dict, Iterable, Optional, Union, cast

from samcli.lib.iac.plugins_interfaces import Stack
from samcli.lib.intrinsic_resolver.intrinsic_property_resolver import IntrinsicResolver
from samcli.lib.intrinsic_resolver.intrinsics_symbol_table import IntrinsicsSymbolTable
from samcli.lib.package.ecr_utils import is_ecr_url
from samcli.lib.samlib.resource_metadata_normalizer import ResourceMetadataNormalizer
from samcli.lib.samlib.wrapper import SamTranslatorWrapper
from samcli.lib.utils.resources import (
    AWS_LAMBDA_FUNCTION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_FUNCTION,
    AWS_SERVERLESS_LAYERVERSION,
)

LOG = logging.getLogger(__name__)


class SamBaseProvider:
    """
    Base class for SAM Template providers
    """

    DEFAULT_CODEURI = "."

    CODE_PROPERTY_KEYS = {
        AWS_LAMBDA_FUNCTION: "Code",
        AWS_SERVERLESS_FUNCTION: "CodeUri",
        AWS_LAMBDA_LAYERVERSION: "Content",
        AWS_SERVERLESS_LAYERVERSION: "ContentUri",
    }

    IMAGE_PROPERTY_KEYS = {
        AWS_LAMBDA_FUNCTION: "Code",
        AWS_SERVERLESS_FUNCTION: "ImageUri",
    }

    def get(self, name: str) -> Optional[Any]:
        """
        Given name of the function, this method must return the Function object

        :param string name: Name of the function
        :return Function: namedtuple containing the Function information
        """
        raise NotImplementedError("not implemented")

    def get_all(self) -> Iterable:
        """
        Yields all the Lambda functions available in the provider.

        :yields Function: namedtuple containing the function information
        """
        raise NotImplementedError("not implemented")

    @staticmethod
    def _extract_codeuri(resource_properties: Dict, code_property_key: str) -> str:
        """
        Extracts the Function/Layer code path from the Resource Properties

        Parameters
        ----------
        resource_properties dict
            Dictionary representing the Properties of the Resource
        code_property_key str
            Property Key of the code on the Resource

        Returns
        -------
        str
            Representing the local code path
        """
        codeuri = resource_properties.get(code_property_key, SamBaseProvider.DEFAULT_CODEURI)

        if isinstance(codeuri, dict):
            return SamBaseProvider.DEFAULT_CODEURI

        return cast(str, codeuri)

    @staticmethod
    def _is_s3_location(location: Optional[Union[str, Dict]]) -> bool:
        """
        the input could be:
        - CodeUri of Serverless::Function
        - Code of Lambda::Function
        - ContentUri of Serverless::LayerVersion
        - Content of Lambda::LayerVersion
        """
        return (isinstance(location, dict) and ("S3Bucket" in location or "Bucket" in location)) or (
            isinstance(location, str) and location.startswith("s3://")
        )

    @staticmethod
    def _is_ecr_uri(location: Optional[Union[str, Dict]]) -> bool:
        """
        the input could be:
        - ImageUri of Serverless::Function
        - Code of Lambda::Function
        """
        return location is not None and is_ecr_url(
            str(location.get("ImageUri", "")) if isinstance(location, dict) else location
        )

    @staticmethod
    def _warn_code_extraction(resource_type: str, resource_name: str, code_property: str) -> None:
        LOG.warning(
            "The resource %s '%s' has specified S3 location for %s. "
            "It will not be built and SAM CLI does not support invoking it locally.",
            resource_type,
            resource_name,
            code_property,
        )

    @staticmethod
    def _warn_imageuri_extraction(resource_type: str, resource_name: str, image_property: str) -> None:
        LOG.warning(
            "The resource %s '%s' has specified ECR registry image for %s. "
            "It will not be built and SAM CLI does not support invoking it locally.",
            resource_type,
            resource_name,
            image_property,
        )

    @staticmethod
    def _extract_lambda_function_imageuri(resource_properties: Dict, code_property_key: str) -> Optional[str]:
        """
        Extracts the Lambda Function ImageUri from the Resource Properties

        Parameters
        ----------
        resource_properties dict
            Dictionary representing the Properties of the Resource
        code_property_key str
            Property Key of the code on the Resource

        Returns
        -------
        str
            Representing the local imageuri
        """
        return cast(Optional[str], resource_properties.get(code_property_key, dict()).get("ImageUri", None))

    @staticmethod
    def _extract_sam_function_imageuri(resource_properties: Dict, code_property_key: str) -> Optional[str]:
        """
        Extracts the Serverless Function ImageUri from the Resource Properties

        Parameters
        ----------
        resource_properties dict
            Dictionary representing the Properties of the Resource
        code_property_key str
            Property Key of the code on the Resource

        Returns
        -------
        str
            Representing the local imageuri
        """
        return resource_properties.get(code_property_key, None)

    @staticmethod
    def get_template(template_dict: Dict, parameter_overrides: Optional[Dict[str, str]] = None) -> Dict:
        """
        Given a SAM template dictionary, return a cleaned copy of the template where SAM plugins have been run
        and parameter values have been substituted.

        Parameters
        ----------
        template_dict : dict
            unprocessed SAM template dictionary

        parameter_overrides: dict
            Optional dictionary of values for template parameters

        Returns
        -------
        dict
            Processed SAM template
        """
        template_dict = template_dict or {}
        parameters_values = SamBaseProvider._get_parameter_values(template_dict, parameter_overrides)
        if template_dict:
            template_dict = SamTranslatorWrapper(template_dict, parameter_values=parameters_values).run_plugins()
        ResourceMetadataNormalizer.normalize(template_dict)

        resolver = IntrinsicResolver(
            template=template_dict,
            symbol_resolver=IntrinsicsSymbolTable(logical_id_translator=parameters_values, template=template_dict),
        )
        template_dict = resolver.resolve_template(ignore_errors=True)
        return template_dict

    @staticmethod
    def get_resolved_template_dict(
        template_dict: Stack,
        parameter_overrides: Optional[Dict[str, str]] = None,
        normalize_resource_metadata: bool = True,
    ) -> Stack:
        """
        Given a SAM template dictionary, return a cleaned copy of the template where SAM plugins have been run
        and parameter values have been substituted.
        Parameters
        ----------
        template_dict : dict
            unprocessed SAM template dictionary
        parameter_overrides: dict
            Optional dictionary of values for template parameters
        normalize_resource_metadata: bool
            flag to normalize resource metadata or not; For package and deploy, we don't need to normalize resource
            metadata, which usually exists in a CDK-synthed template and is used for build and local testing
        Returns
        -------
        dict
            Processed SAM template
            :param template_dict:
            :param parameter_overrides:
            :param normalize_resource_metadata:
        """
        template_dict = template_dict or Stack()
        parameters_values = SamBaseProvider._get_parameter_values(template_dict, parameter_overrides)
        if template_dict:
            template_dict = SamTranslatorWrapper(template_dict, parameter_values=parameters_values).run_plugins()
        if normalize_resource_metadata:
            ResourceMetadataNormalizer.normalize(template_dict)

        resolver = IntrinsicResolver(
            template=template_dict,
            symbol_resolver=IntrinsicsSymbolTable(logical_id_translator=parameters_values, template=template_dict),
        )
        template_dict = resolver.resolve_template(ignore_errors=True)
        return template_dict

    @staticmethod
    def _get_parameter_values(template_dict: Any, parameter_overrides: Optional[Dict]) -> Dict:
        """
        Construct a final list of values for CloudFormation template parameters based on user-supplied values,
        default values provided in template, and sane defaults for pseudo-parameters.

        Parameters
        ----------
        template_dict : dict
            SAM template dictionary

        parameter_overrides : dict
            User-supplied values for CloudFormation template parameters

        Returns
        -------
        dict
            Values for template parameters to substitute in template with
        """

        default_values = SamBaseProvider._get_default_parameter_values(template_dict)

        # NOTE: Ordering of following statements is important. It makes sure that any user-supplied values
        # override the defaults
        parameter_values = {}
        parameter_values.update(IntrinsicsSymbolTable.DEFAULT_PSEUDO_PARAM_VALUES)
        parameter_values.update(default_values)
        parameter_values.update(parameter_overrides or {})

        return parameter_values

    @staticmethod
    def _get_default_parameter_values(sam_template: Dict) -> Dict:
        """
        Method to read default values for template parameters and return it
        Example:
        If the template contains the following parameters defined
        Parameters:
            Param1:
                Type: String
                Default: default_value1
            Param2:
                Type: String
                Default: default_value2

        then, this method will grab default value for Param1 and return the following result:
        {
            Param1: "default_value1",
            Param2: "default_value2"
        }
        :param dict sam_template: SAM template
        :return dict: Default values for parameters
        """

        default_values: Dict = {}

        parameter_definition = sam_template.get("Parameters", None)
        if not parameter_definition or not isinstance(parameter_definition, dict):
            LOG.debug("No Parameters detected in the template")
            return default_values

        for param_name, value in parameter_definition.items():
            if isinstance(value, dict) and "Default" in value:
                default_values[param_name] = value["Default"]

        LOG.debug("Collected default values for parameters: %s", default_values)
        return default_values



"""
Enums for Resources and their Location Properties, along with utility functions
"""

from collections import defaultdict

# Lambda
AWS_SERVERLESS_FUNCTION = "AWS::Serverless::Function"
AWS_SERVERLESS_LAYERVERSION = "AWS::Serverless::LayerVersion"

AWS_LAMBDA_FUNCTION = "AWS::Lambda::Function"
AWS_LAMBDA_LAYERVERSION = "AWS::Lambda::LayerVersion"

# APIGW
AWS_SERVERLESS_API = "AWS::Serverless::Api"
AWS_SERVERLESS_HTTPAPI = "AWS::Serverless::HttpApi"

AWS_APIGATEWAY_RESTAPI = "AWS::ApiGateway::RestApi"
AWS_APIGATEWAY_STAGE = "AWS::ApiGateway::Stage"
AWS_APIGATEWAY_RESOURCE = "AWS::ApiGateway::Resource"
AWS_APIGATEWAY_METHOD = "AWS::ApiGateway::Method"
AWS_APIGATEWAY_DEPLOYMENT = "AWS::ApiGateway::Deployment"
AWS_APIGATEWAY_BASE_PATH_MAPPING = "AWS::ApiGateway::BasePathMapping"
AWS_APIGATWAY_DOMAIN_NAME = "AWS::ApiGateway::DomainName"

AWS_APIGATEWAY_V2_API = "AWS::ApiGatewayV2::Api"
AWS_APIGATEWAY_V2_INTEGRATION = "AWS::ApiGatewayV2::Integration"
AWS_APIGATEWAY_V2_ROUTE = "AWS::ApiGatewayV2::Route"
AWS_APIGATEWAY_V2_STAGE = "AWS::ApiGatewayV2::Stage"
AWS_APIGATEWAY_v2_BASE_PATH_MAPPING = "AWS::ApiGatewayV2::ApiMapping"
AWS_APIGATEWAY_V2_DOMAIN_NAME = "AWS::ApiGatewayV2::DomainName"

# SFN
AWS_SERVERLESS_STATEMACHINE = "AWS::Serverless::StateMachine"

AWS_STEPFUNCTIONS_STATEMACHINE = "AWS::StepFunctions::StateMachine"

# Others
AWS_SERVERLESS_APPLICATION = "AWS::Serverless::Application"

AWS_SERVERLESSREPO_APPLICATION = "AWS::ServerlessRepo::Application"
AWS_APPSYNC_GRAPHQLSCHEMA = "AWS::AppSync::GraphQLSchema"
AWS_APPSYNC_RESOLVER = "AWS::AppSync::Resolver"
AWS_APPSYNC_FUNCTIONCONFIGURATION = "AWS::AppSync::FunctionConfiguration"
AWS_ELASTICBEANSTALK_APPLICATIONVERSION = "AWS::ElasticBeanstalk::ApplicationVersion"
AWS_CLOUDFORMATION_MODULEVERSION = "AWS::CloudFormation::ModuleVersion"
AWS_CLOUDFORMATION_RESOURCEVERSION = "AWS::CloudFormation::ResourceVersion"
AWS_CLOUDFORMATION_STACK = "AWS::CloudFormation::Stack"
AWS_CLOUDFORMATION_STACKSET = "AWS::CloudFormation::StackSet"
AWS_GLUE_JOB = "AWS::Glue::Job"
AWS_SQS_QUEUE = "AWS::SQS::Queue"
AWS_KINESIS_STREAM = "AWS::Kinesis::Stream"
AWS_SERVERLESS_STATEMACHINE = "AWS::Serverless::StateMachine"
AWS_STEPFUNCTIONS_STATEMACHINE = "AWS::StepFunctions::StateMachine"
AWS_ECR_REPOSITORY = "AWS::ECR::Repository"
AWS_APPLICATION_INSIGHTS = "AWS::ApplicationInsights::Application"
AWS_RESOURCE_GROUP = "AWS::ResourceGroups::Group"

METADATA_WITH_LOCAL_PATHS = {AWS_SERVERLESSREPO_APPLICATION: ["LicenseUrl", "ReadmeUrl"]}

RESOURCES_WITH_LOCAL_PATHS = {
    AWS_SERVERLESS_FUNCTION: ["CodeUri"],
    AWS_SERVERLESS_API: ["DefinitionUri"],
    AWS_SERVERLESS_HTTPAPI: ["DefinitionUri"],
    AWS_SERVERLESS_STATEMACHINE: ["DefinitionUri"],
    AWS_APPSYNC_GRAPHQLSCHEMA: ["DefinitionS3Location"],
    AWS_APPSYNC_RESOLVER: ["RequestMappingTemplateS3Location", "ResponseMappingTemplateS3Location", "CodeS3Location"],
    AWS_APPSYNC_FUNCTIONCONFIGURATION: [
        "RequestMappingTemplateS3Location",
        "ResponseMappingTemplateS3Location",
        "CodeS3Location",
    ],
    AWS_LAMBDA_FUNCTION: ["Code"],
    AWS_APIGATEWAY_RESTAPI: ["BodyS3Location"],
    AWS_APIGATEWAY_V2_API: ["BodyS3Location"],
    AWS_ELASTICBEANSTALK_APPLICATIONVERSION: ["SourceBundle"],
    AWS_CLOUDFORMATION_MODULEVERSION: ["ModulePackage"],
    AWS_CLOUDFORMATION_RESOURCEVERSION: ["SchemaHandlerPackage"],
    AWS_CLOUDFORMATION_STACK: ["TemplateURL"],
    AWS_CLOUDFORMATION_STACKSET: ["TemplateURL"],
    AWS_SERVERLESS_APPLICATION: ["Location"],
    AWS_LAMBDA_LAYERVERSION: ["Content"],
    AWS_SERVERLESS_LAYERVERSION: ["ContentUri"],
    AWS_GLUE_JOB: ["Command.ScriptLocation"],
    AWS_STEPFUNCTIONS_STATEMACHINE: ["DefinitionS3Location"],
}

RESOURCES_WITH_IMAGE_COMPONENT = {
    AWS_SERVERLESS_FUNCTION: ["ImageUri"],
    AWS_LAMBDA_FUNCTION: ["Code.ImageUri"],
    AWS_ECR_REPOSITORY: ["RepositoryName"],
}

NESTED_STACKS_RESOURCES = {
    AWS_SERVERLESS_APPLICATION: "Location",
    AWS_CLOUDFORMATION_STACK: "TemplateURL",
}

LAMBDA_LOCAL_RESOURCES = [
    AWS_LAMBDA_FUNCTION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_FUNCTION,
    AWS_SERVERLESS_LAYERVERSION,
]

CODE_SYNCABLE_RESOURCES = [
    AWS_SERVERLESS_FUNCTION,
    AWS_LAMBDA_FUNCTION,
    AWS_SERVERLESS_LAYERVERSION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_API,
    AWS_APIGATEWAY_RESTAPI,
    AWS_SERVERLESS_HTTPAPI,
    AWS_APIGATEWAY_V2_API,
    AWS_SERVERLESS_STATEMACHINE,
    AWS_STEPFUNCTIONS_STATEMACHINE,
]

SYNCABLE_STACK_RESOURCES = [AWS_SERVERLESS_APPLICATION, AWS_CLOUDFORMATION_STACK]

AWS_LAMBDA_FUNCTION_URL = "AWS::Lambda::Url"


def get_packageable_resource_paths():
    """
    Resource Types with respective Locations that are package-able.

    Returns
    ------
    _resource_property_dict : Dict
        Resource Dictionary containing packageable resource types and their locations as a list.
    """
    _resource_property_dict = defaultdict(list)
    for _dict in (METADATA_WITH_LOCAL_PATHS, RESOURCES_WITH_LOCAL_PATHS, RESOURCES_WITH_IMAGE_COMPONENT):
        for key, value in _dict.items():
            # Only add values to the list if they are different, same property name could be used with the resource
            # to package to different locations.
            if value not in _resource_property_dict.get(key, []):
                _resource_property_dict[key].append(value)

    return _resource_property_dict


def resources_generator():
    """
    Generator to yield set of resources and their locations that are supported for package operations

    Yields
    ------
    resource : Dict
        The resource dictionary
    location : str
        The location of the resource
    """
    for resource, location_list in get_packageable_resource_paths().items():
        for locations in location_list:
            for location in locations:
                yield resource, location



"""
Utilities to manipulate template
"""
import itertools
import os
import pathlib

import jmespath
import yaml
from botocore.utils import set_value_from_jmespath

from samcli.commands.exceptions import UserException
from samcli.lib.samlib.resource_metadata_normalizer import ASSET_PATH_METADATA_KEY, ResourceMetadataNormalizer
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import (
    AWS_LAMBDA_FUNCTION,
    AWS_SERVERLESS_FUNCTION,
    METADATA_WITH_LOCAL_PATHS,
    RESOURCES_WITH_LOCAL_PATHS,
    get_packageable_resource_paths,
)
from samcli.yamlhelper import yaml_dump, yaml_parse


class TemplateNotFoundException(UserException):
    pass


class TemplateFailedParsingException(UserException):
    pass


def get_template_data(template_file):
    """
    Read the template file, parse it as JSON/YAML and return the template as a dictionary.

    Parameters
    ----------
    template_file : string
        Path to the template to read

    Returns
    -------
    Template data as a dictionary
    """

    if not pathlib.Path(template_file).exists():
        raise TemplateNotFoundException("Template file not found at {}".format(template_file))

    with open(template_file, "r", encoding="utf-8") as fp:
        try:
            return yaml_parse(fp.read())
        except (ValueError, yaml.YAMLError) as ex:
            raise TemplateFailedParsingException("Failed to parse template: {}".format(str(ex))) from ex


def move_template(src_template_path, dest_template_path, template_dict):
    """
    Move the SAM/CloudFormation template from ``src_template_path`` to ``dest_template_path``. For convenience, this
    method accepts a dictionary of template data ``template_dict`` that will be written to the destination instead of
    reading from the source file.

    SAM/CloudFormation template can contain certain properties whose value is a relative path to a local file/folder.
    This path is always relative to the template's location. Before writing the template to ``dest_template_path`,
    we will update these paths to be relative to the new location.

    This methods updates resource properties supported by ``aws cloudformation package`` command:
    https://docs.aws.amazon.com/cli/latest/reference/cloudformation/package.html

    You must use this method if you are reading a template from one location, modifying it, and writing it back to a
    different location.

    Parameters
    ----------
    src_template_path : str
        Path to the original location of the template

    dest_template_path : str
        Path to the destination location where updated template should be written to

    template_dict : dict
        Dictionary containing template contents. This dictionary will be updated & written to ``dest`` location.
    """

    original_root = os.path.dirname(src_template_path)
    new_root = os.path.dirname(dest_template_path)

    # Next up, we will be writing the template to a different location. Before doing so, we should
    # update any relative paths in the template to be relative to the new location.
    modified_template = _update_relative_paths(template_dict, original_root, new_root)

    # if a stack only has image functions, the directory for that directory won't be created.
    # here we make sure the directory the destination template file to write to exists.
    os.makedirs(os.path.dirname(dest_template_path), exist_ok=True)
    with open(dest_template_path, "w") as fp:
        fp.write(yaml_dump(modified_template))


def _update_relative_paths(template_dict, original_root, new_root):
    """
    SAM/CloudFormation template can contain certain properties whose value is a relative path to a local file/folder.
    This path is usually relative to the template's location. If the template is being moved from original location
    ``original_root`` to new location ``new_root``, use this method to update these paths to be
    relative to ``new_root``.

    After this method is complete, it is safe to write the template to ``new_root`` without
    breaking any relative paths.

    This methods updates resource properties supported by ``aws cloudformation package`` command:
    https://docs.aws.amazon.com/cli/latest/reference/cloudformation/package.html

    If a property is either an absolute path or a S3 URI, this method will not update them.


    Parameters
    ----------
    template_dict : dict
        Dictionary containing template contents. This dictionary will be updated & written to ``dest`` location.

    original_root : str
        Path to the directory where all paths were originally set relative to. This is usually the directory
        containing the template originally

    new_root : str
        Path to the new directory that all paths set relative to after this method completes.

    Returns
    -------
    Updated dictionary

    """

    for resource_type, properties in template_dict.get("Metadata", {}).items():

        if resource_type not in METADATA_WITH_LOCAL_PATHS:
            # Unknown resource. Skipping
            continue

        for path_prop_name in METADATA_WITH_LOCAL_PATHS[resource_type]:
            path = properties.get(path_prop_name)

            updated_path = _resolve_relative_to(path, original_root, new_root)
            if not updated_path:
                # This path does not need to get updated
                continue

            properties[path_prop_name] = updated_path

    for _, resource in template_dict.get("Resources", {}).items():
        resource_type = resource.get("Type")

        if resource_type not in RESOURCES_WITH_LOCAL_PATHS:
            # Unknown resource. Skipping
            continue

        for path_prop_name in RESOURCES_WITH_LOCAL_PATHS[resource_type]:
            properties = resource.get("Properties", {})

            if (
                resource_type in [AWS_SERVERLESS_FUNCTION, AWS_LAMBDA_FUNCTION]
                and properties.get("PackageType", ZIP) == IMAGE
            ):
                continue

            path = jmespath.search(path_prop_name, properties)
            updated_path = _resolve_relative_to(path, original_root, new_root)

            if not updated_path:
                # This path does not need to get updated
                continue

            set_value_from_jmespath(properties, path_prop_name, updated_path)

        metadata = resource.get("Metadata", {})
        if ASSET_PATH_METADATA_KEY in metadata:
            path = metadata.get(ASSET_PATH_METADATA_KEY, "")
            updated_path = _resolve_relative_to(path, original_root, new_root)
            if not updated_path:
                # This path does not need to get updated
                continue
            metadata[ASSET_PATH_METADATA_KEY] = updated_path

    # AWS::Includes can be anywhere within the template dictionary. Hence we need to recurse through the
    # dictionary in a separate method to find and update relative paths in there
    template_dict = _update_aws_include_relative_path(template_dict, original_root, new_root)

    return template_dict


def _update_aws_include_relative_path(template_dict, original_root, new_root):
    """
    Update relative paths in "AWS::Include" directive. This directive can be present at any part of the template,
    and not just within resources.
    """

    for key, val in template_dict.items():
        if key == "Fn::Transform":
            if isinstance(val, dict) and val.get("Name") == "AWS::Include":
                path = val.get("Parameters", {}).get("Location", {})
                updated_path = _resolve_relative_to(path, original_root, new_root)
                if not updated_path:
                    # This path does not need to get updated
                    continue

                val["Parameters"]["Location"] = updated_path

        # Recurse through all dictionary values
        elif isinstance(val, dict):
            _update_aws_include_relative_path(val, original_root, new_root)
        elif isinstance(val, list):
            for item in val:
                if isinstance(item, dict):
                    _update_aws_include_relative_path(item, original_root, new_root)

    return template_dict


def _resolve_relative_to(path, original_root, new_root):
    """
    If the given ``path`` is a relative path, then assume it is relative to ``original_root``. This method will
    update the path to be resolve it relative to ``new_root`` and return.

    Examples
    -------
        # Assume a file called template.txt at location /tmp/original/root/template.txt expressed as relative path
        # We are trying to update it to be relative to /tmp/new/root instead of the /tmp/original/root
        >>> result = _resolve_relative_to("template.txt",  \
                                          "/tmp/original/root", \
                                          "/tmp/new/root")
        >>> result
        ../../original/root/template.txt

    Returns
    -------
    Updated path if the given path is a relative path. None, if the path is not a relative path.
    """

    if (
        not isinstance(path, str)
        or path.startswith("s3://")
        or path.startswith("http://")
        or path.startswith("https://")
        or os.path.isabs(path)
    ):
        # Value is definitely NOT a relative path. It is either a S3 URi or Absolute path or not a string at all
        return None

    # Value is definitely a relative path. Change it relative to the destination directory
    return os.path.relpath(
        # Resolve the paths to take care of symlinks
        os.path.normpath(os.path.join(pathlib.Path(original_root).resolve(), path)),
        pathlib.Path(new_root).resolve(),  # Absolute original path w.r.t ``original_root``
    )  # Resolve the original path with respect to ``new_root``


def get_template_parameters(template_file):
    """
    Get Parameters from a template file.

    Parameters
    ----------
    template_file : string
        Path to the template to read

    Returns
    -------
    Template Parameters as a dictionary
    """
    template_dict = get_template_data(template_file=template_file)
    ResourceMetadataNormalizer.normalize(template_dict, True)
    return template_dict.get("Parameters", dict())


def get_template_artifacts_format(template_file):
    """
    Get a list of template artifact formats based on PackageType wherever the underlying resource
    have the actual need to be packaged.
    :param template_file:
    :return: list of artifact formats
    """

    template_dict = get_template_data(template_file=template_file)

    # Get a list of Resources where the artifacts format matter for packaging.
    packageable_resources = get_packageable_resource_paths()

    artifacts = []
    for _, resource in template_dict.get("Resources", {}).items():
        # First check if the resources are part of package-able resource types.
        if resource.get("Type") in packageable_resources.keys():
            # Flatten list of locations per resource type.
            locations = list(itertools.chain(*packageable_resources.get(resource.get("Type"))))
            for location in locations:
                properties = resource.get("Properties", {})
                # Search for package-able location within resource properties.
                if jmespath.search(location, properties):
                    artifacts.append(properties.get("PackageType", ZIP))

    return artifacts


def get_template_function_resource_ids(template_file, artifact):
    """
    Get a list of function logical ids from template file.
    Function resource types include
        AWS::Lambda::Function
        AWS::Serverless::Function
    :param template_file: template file location.
    :param artifact: artifact of type IMAGE or ZIP
    :return: list of artifact formats
    """

    template_dict = get_template_data(template_file=template_file)
    _function_resource_ids = []
    for resource_id, resource in template_dict.get("Resources", {}).items():
        if resource.get("Properties", {}).get("PackageType", ZIP) == artifact and resource.get("Type") in [
            AWS_SERVERLESS_FUNCTION,
            AWS_LAMBDA_FUNCTION,
        ]:
            _function_resource_ids.append(resource_id)
    return _function_resource_ids



"""
Module used for detecting whether SAMCLI is executed in a CI/CD environment.
"""

import os
from enum import Enum, auto
from typing import Callable, Dict, Mapping, Optional, Union


class CICDPlatform(Enum):
    Jenkins = auto()
    GitLab = auto()
    GitHubAction = auto()
    TravisCI = auto()
    CircleCI = auto()
    AWSCodeBuild = auto()
    TeamCity = auto()
    Bamboo = auto()
    Buddy = auto()
    CodeShip = auto()
    Semaphore = auto()
    Appveyor = auto()
    # make sure Unknown is at the bottom, it is the fallback.
    Other = auto()


def _is_codeship(environ: Mapping) -> bool:
    """
    Use environ to determine whether it is running in CodeShip.
    According to the doc,
    https://docs.cloudbees.com/docs/cloudbees-codeship/latest/basic-builds-and-configuration/set-environment-variables
    > CI_NAME                # Always CodeShip. Ex: codeship

    to handle both "CodeShip" and "codeship," here the string is converted to lower case first.

    Parameters
    ----------
    environ

    Returns
    -------
    bool
        whether the env is CodeShip
    """
    ci_name: str = environ.get("CI_NAME", "").lower()
    return ci_name == "codeship"


def _is_jenkins(environ: Mapping) -> bool:
    """
    Use environ to determine whether it is running in Jenkins.
    According to the doc,
    https://www.jenkins.io/doc/book/pipeline/jenkinsfile/#working-with-your-jenkinsfile
    > BUILD_TAG
    >   String of jenkins-${JOB_NAME}-${BUILD_NUMBER}.
    > ...
    > JENKINS_URL
    >   Full URL of Jenkins, such as https://example.com:port/jenkins/
    >   (NOTE: only available if Jenkins URL set in "System Configuration")

    Here firstly check JENKINS_URL's presence, if not, then fallback to check BUILD_TAG starts with "jenkins"
    """
    return "JENKINS_URL" in environ or environ.get("BUILD_TAG", "").startswith("jenkins-")


_ENV_VAR_OR_CALLABLE_BY_PLATFORM: Dict[CICDPlatform, Union[str, Callable[[Mapping], bool]]] = {
    CICDPlatform.Jenkins: _is_jenkins,
    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    CICDPlatform.GitLab: "GITLAB_CI",
    # https://docs.github.com/en/actions/reference/environment-variables
    CICDPlatform.GitHubAction: "GITHUB_ACTION",
    # https://docs.travis-ci.com/user/environment-variables/
    CICDPlatform.TravisCI: "TRAVIS",
    # https://circleci.com/docs/2.0/env-vars/
    CICDPlatform.CircleCI: "CIRCLECI",
    # https://docs.aws.amazon.com/codebuild/latest/userguide/build-env-ref-env-vars.html
    CICDPlatform.AWSCodeBuild: "CODEBUILD_BUILD_ID",
    # https://www.jetbrains.com/help/teamcity/predefined-build-parameters.html
    CICDPlatform.TeamCity: "TEAMCITY_VERSION",
    # https://confluence.atlassian.com/bamboo/bamboo-variables-289277087.html
    CICDPlatform.Bamboo: "bamboo_buildNumber",
    # https://buddy.works/docs/pipelines/environment-variables
    CICDPlatform.Buddy: "BUDDY",
    CICDPlatform.CodeShip: _is_codeship,
    # https://docs.semaphoreci.com/ci-cd-environment/environment-variables/
    CICDPlatform.Semaphore: "SEMAPHORE",
    # https://www.appveyor.com/docs/environment-variables/
    CICDPlatform.Appveyor: "APPVEYOR",
    CICDPlatform.Other: "CI",
}


def _is_cicd_platform(cicd_platform: CICDPlatform, environ: Mapping) -> bool:
    """
    Check whether sam-cli run in a particular CI/CD platform based on certain environment variables.

    Parameters
    ----------
    cicd_platform
        an enum CICDPlatform object indicating which CI/CD platform  to check against.
    environ
        the mapping to look for environment variables, for example, os.environ.

    Returns
    -------
    bool
        A boolean indicating whether there are environment variables matching the cicd_platform.
    """
    env_var_or_callable = _ENV_VAR_OR_CALLABLE_BY_PLATFORM[cicd_platform]
    if isinstance(env_var_or_callable, str):
        return env_var_or_callable in environ

    # it is a callable, use the return value
    return env_var_or_callable(environ)


class CICDDetector:
    _cicd_platform: Optional[CICDPlatform]

    def __init__(self):
        try:
            self._cicd_platform: Optional[CICDPlatform] = next(
                cicd_platform for cicd_platform in CICDPlatform if _is_cicd_platform(cicd_platform, os.environ)
            )
        except StopIteration:
            self._cicd_platform = None

    def platform(self) -> Optional[CICDPlatform]:
        """
        Identify which CICD platform SAM CLI is running in.
        Returns
        -------
        CICDPlatform
            an optional CICDPlatform enum indicating the CICD platform.
        """
        return self._cicd_platform



"""
Utilities for Deploy
"""

from enum import Enum

from samcli.lib.utils.colors import Colored


class DeployColor:
    def __init__(self):
        self._color = Colored()
        self.changeset_color_map = {"Add": "green", "Modify": "yellow", "Remove": "red"}
        self.status_color_map = {
            "CREATE_COMPLETE": "green",
            "CREATE_FAILED": "red",
            "CREATE_IN_PROGRESS": "yellow",
            "DELETE_COMPLETE": "green",
            "DELETE_FAILED": "red",
            "DELETE_IN_PROGRESS": "red",
            "REVIEW_IN_PROGRESS": "yellow",
            "ROLLBACK_COMPLETE": "red",
            "ROLLBACK_IN_PROGRESS": "red",
            "UPDATE_COMPLETE": "green",
            "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS": "yellow",
            "UPDATE_IN_PROGRESS": "yellow",
            "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS": "red",
            "UPDATE_ROLLBACK_FAILED": "red",
            "UPDATE_ROLLBACK_IN_PROGRESS": "red",
        }

    def get_stack_events_status_color(self, status):
        return self.status_color_map.get(status, "yellow")

    def get_changeset_action_color(self, action):
        return self.changeset_color_map.get(action, "yellow")


class FailureMode(Enum):
    ROLLBACK = "ROLLBACK"
    DELETE = "DELETE"
    DO_NOTHING = "DO_NOTHING"

    def __str__(self):
        return self.value



"""
Terraform resource enrichment

This module populates the values required for each of the Lambda resources
"""
import json
import logging
import os
import re
from json.decoder import JSONDecodeError
from subprocess import CalledProcessError, run
from typing import Dict, List, Tuple

from samcli.hook_packages.terraform.hooks.prepare.constants import (
    CFN_CODE_PROPERTIES,
    SAM_METADATA_RESOURCE_NAME_ATTRIBUTE,
)
from samcli.hook_packages.terraform.hooks.prepare.exceptions import InvalidSamMetadataPropertiesException
from samcli.hook_packages.terraform.hooks.prepare.makefile_generator import (
    generate_makefile,
    generate_makefile_rule_for_lambda_resource,
)
from samcli.hook_packages.terraform.hooks.prepare.resource_linking import _resolve_resource_attribute
from samcli.hook_packages.terraform.hooks.prepare.types import SamMetadataResource
from samcli.hook_packages.terraform.lib.utils import (
    _calculate_configuration_attribute_value_hash,
    build_cfn_logical_id,
    get_sam_metadata_planned_resource_value_attribute,
)
from samcli.lib.hook.exceptions import PrepareHookException
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import AWS_LAMBDA_FUNCTION as CFN_AWS_LAMBDA_FUNCTION
from samcli.lib.utils.resources import AWS_LAMBDA_LAYERVERSION as CFN_AWS_LAMBDA_LAYER_VERSION

SAM_METADATA_DOCKER_TAG_ATTRIBUTE = "docker_tag"
SAM_METADATA_DOCKER_BUILD_ARGS_ATTRIBUTE = "docker_build_args"
SAM_METADATA_DOCKER_FILE_ATTRIBUTE = "docker_file"
SAM_METADATA_RESOURCE_TYPE_ATTRIBUTE = "resource_type"

# check for python 3, 3.7 or above
# regex: search for 'Python', whitespace, '3.', digits 7-9 or 2+ digits, any digit or '.' 0+ times
PYTHON_VERSION_REGEX = re.compile(r"Python\s*3.([7-9]|\d{2,})[\d.]*")

LOG = logging.getLogger(__name__)


def enrich_resources_and_generate_makefile(
    sam_metadata_resources: List[SamMetadataResource],
    cfn_resources: Dict[str, Dict],
    output_directory_path: str,
    terraform_application_dir: str,
    lambda_resources_to_code_map: Dict,
) -> None:
    """
    Use the sam metadata resources to enrich the mapped resources and to create a Makefile with a rule for
    each lambda resource to be built.

    Parameters
    ----------
    sam_metadata_resources: List[SamMetadataResource]
        The list of sam metadata resources defined in the terraform project.
    cfn_resources: dict
        CloudFormation resources
    output_directory_path: str
        the output directory path to write the generated metadata and makefile
    terraform_application_dir: str
        the terraform project root directory
    lambda_resources_to_code_map: Dict
        The map between lambda resources code path, and lambda resources logical ids
    """

    python_command_name = _get_python_command_name()

    resources_types_enrichment_functions = {
        "ZIP_LAMBDA_FUNCTION": _enrich_zip_lambda_function,
        "IMAGE_LAMBDA_FUNCTION": _enrich_image_lambda_function,
        "LAMBDA_LAYER": _enrich_lambda_layer,
    }

    makefile_rules = []
    for sam_metadata_resource in sam_metadata_resources:
        # enrich resource
        resource_type = get_sam_metadata_planned_resource_value_attribute(
            sam_metadata_resource.resource, SAM_METADATA_RESOURCE_TYPE_ATTRIBUTE
        )
        sam_metadata_resource_address = sam_metadata_resource.resource.get("address")
        enrichment_function = resources_types_enrichment_functions.get(resource_type)
        if enrichment_function is None:
            raise InvalidSamMetadataPropertiesException(
                f"The resource type {resource_type} found in the sam metadata resource "
                f"{sam_metadata_resource_address} is not a correct resource type. The resource type should be one "
                f"of these values {resources_types_enrichment_functions.keys()}"
            )

        lambda_resources = _get_relevant_cfn_resource(
            sam_metadata_resource, cfn_resources, lambda_resources_to_code_map
        )
        for cfn_resource, logical_id in lambda_resources:
            enrichment_function(
                sam_metadata_resource.resource,
                cfn_resource,
                logical_id,
                terraform_application_dir,
                output_directory_path,
            )

            # get makefile rule for resource
            makefile_rule = generate_makefile_rule_for_lambda_resource(
                sam_metadata_resource, logical_id, terraform_application_dir, python_command_name, output_directory_path
            )
            makefile_rules.append(makefile_rule)

    # generate makefile
    LOG.debug("Generate Makefile in %s", output_directory_path)
    generate_makefile(makefile_rules, output_directory_path)


def _enrich_zip_lambda_function(
    sam_metadata_resource: Dict,
    cfn_lambda_function: Dict,
    cfn_lambda_function_logical_id: str,
    terraform_application_dir: str,
    output_directory_path: str,
):
    """
    Use the sam metadata resources to enrich the zip lambda function.

    Parameters
    ----------
    sam_metadata_resource: Dict
        The sam metadata resource properties
    cfn_lambda_function: dict
        CloudFormation lambda function to be enriched
    cfn_lambda_function_logical_id: str
        the cloudFormation lambda function to be enriched logical id.
    output_directory_path: str
        the output directory path to write the generated metadata and makefile
    terraform_application_dir: str
        the terraform project root directory
    """
    sam_metadata_resource_address = sam_metadata_resource.get("address")
    if not sam_metadata_resource_address:
        raise PrepareHookException(
            "Invalid Terraform plan output. The address property should not be null to any terraform resource."
        )

    LOG.debug(
        "Enrich the ZIP lambda function %s using the metadata properties defined in resource %s",
        cfn_lambda_function_logical_id,
        sam_metadata_resource_address,
    )

    _validate_referenced_resource_matches_sam_metadata_type(
        cfn_lambda_function, sam_metadata_resource, sam_metadata_resource_address, ZIP
    )

    cfn_source_code_path = _get_source_code_path(
        sam_metadata_resource,
        sam_metadata_resource_address,
        terraform_application_dir,
        "original_source_code",
        "source_code_property",
        "source code",
    )
    _set_zip_metadata_resources(
        cfn_lambda_function,
        cfn_source_code_path,
        output_directory_path,
        terraform_application_dir,
        CFN_CODE_PROPERTIES[CFN_AWS_LAMBDA_FUNCTION],
    )


def _enrich_image_lambda_function(
    sam_metadata_resource: Dict,
    cfn_lambda_function: Dict,
    cfn_lambda_function_logical_id: str,
    terraform_application_dir: str,
    output_directory_path: str,
):
    """
    Use the sam metadata resources to enrich the image lambda function.

    Parameters
    ----------
    sam_metadata_resource: Dict
        The sam metadata resource properties
    cfn_lambda_function: dict
        CloudFormation lambda function to be enriched
    cfn_lambda_function_logical_id: str
        the cloudFormation lambda function to be enriched logical id.
    output_directory_path: str
        the output directory path to write the generated metadata and makefile
    terraform_application_dir: str
        the terraform project root directory
    """
    sam_metadata_resource_address = sam_metadata_resource.get("address")
    if not sam_metadata_resource_address:
        raise PrepareHookException(
            "Invalid Terraform plan output. The address property should not be null to any terraform resource."
        )
    cfn_resource_properties = cfn_lambda_function.get("Properties", {})

    LOG.debug(
        "Enrich the IMAGE lambda function %s using the metadata properties defined in resource %s",
        cfn_lambda_function_logical_id,
        sam_metadata_resource_address,
    )

    _validate_referenced_resource_matches_sam_metadata_type(
        cfn_lambda_function, sam_metadata_resource, sam_metadata_resource_address, IMAGE
    )

    cfn_docker_context_path = _get_source_code_path(
        sam_metadata_resource,
        sam_metadata_resource_address,
        terraform_application_dir,
        "docker_context",
        "docker_context_property_path",
        "docker context",
    )
    cfn_docker_file = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource, SAM_METADATA_DOCKER_FILE_ATTRIBUTE
    )
    cfn_docker_build_args_string = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource, SAM_METADATA_DOCKER_BUILD_ARGS_ATTRIBUTE
    )
    cfn_docker_build_args = None
    if cfn_docker_build_args_string:
        try:
            LOG.debug("Parse the docker build args %s", cfn_docker_build_args_string)
            cfn_docker_build_args = json.loads(cfn_docker_build_args_string)
            if not isinstance(cfn_docker_build_args, dict):
                raise InvalidSamMetadataPropertiesException(
                    f"The sam metadata resource {sam_metadata_resource_address} should contain a valid json "
                    f"encoded string for the lambda function docker build arguments."
                )
        except JSONDecodeError as exc:
            raise InvalidSamMetadataPropertiesException(
                f"The sam metadata resource {sam_metadata_resource_address} should contain a valid json encoded "
                f"string for the lambda function docker build arguments."
            ) from exc

    cfn_docker_tag = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource, SAM_METADATA_DOCKER_TAG_ATTRIBUTE
    )

    if cfn_resource_properties.get("Code"):
        cfn_resource_properties.pop("Code")

    if not cfn_lambda_function.get("Metadata", {}):
        cfn_lambda_function["Metadata"] = {}
    cfn_lambda_function["Metadata"]["SkipBuild"] = False
    cfn_lambda_function["Metadata"]["DockerContext"] = cfn_docker_context_path
    if cfn_docker_file:
        cfn_lambda_function["Metadata"]["Dockerfile"] = cfn_docker_file
    if cfn_docker_tag:
        cfn_lambda_function["Metadata"]["DockerTag"] = cfn_docker_tag
    if cfn_docker_build_args:
        cfn_lambda_function["Metadata"]["DockerBuildArgs"] = cfn_docker_build_args


def _enrich_lambda_layer(
    sam_metadata_resource: Dict,
    cfn_lambda_layer: Dict,
    cfn_lambda_layer_logical_id: str,
    terraform_application_dir: str,
    output_directory_path: str,
) -> None:
    """
    Use the sam metadata resources to enrich the lambda layer.

    Parameters
    ----------
    sam_metadata_resource: Dict
       The sam metadata resource properties
    cfn_lambda_layer: dict
       CloudFormation lambda layer to be enriched
    cfn_lambda_layer_logical_id: str
       the cloudFormation lambda layer to be enriched logical id.
    output_directory_path: str
       the output directory path to write the generated metadata and makefile
    terraform_application_dir: str
       the terraform project root directory
    """
    sam_metadata_resource_address = sam_metadata_resource.get("address")
    if not sam_metadata_resource_address:
        raise PrepareHookException(
            "Invalid Terraform plan output. The address property should not be null to any terraform resource."
        )
    _validate_referenced_resource_layer_matches_metadata_type(
        cfn_lambda_layer, sam_metadata_resource, sam_metadata_resource_address
    )
    LOG.debug(
        "Enrich the Lambda Layer Version %s using the metadata properties defined in resource %s",
        cfn_lambda_layer_logical_id,
        sam_metadata_resource_address,
    )

    cfn_source_code_path = _get_source_code_path(
        sam_metadata_resource,
        sam_metadata_resource_address,
        terraform_application_dir,
        "original_source_code",
        "source_code_property",
        "source code",
    )

    _set_zip_metadata_resources(
        cfn_lambda_layer,
        cfn_source_code_path,
        output_directory_path,
        terraform_application_dir,
        CFN_CODE_PROPERTIES[CFN_AWS_LAMBDA_LAYER_VERSION],
    )


def _validate_referenced_resource_layer_matches_metadata_type(
    cfn_resource: dict,
    sam_metadata_resource: dict,
    sam_metadata_resource_address: str,
) -> None:
    """
    Validate if the resource that match the resource name provided in the sam metadata resource matches the resource
    type provided in the metadata as well.

    Parameters
    ----------
    cfn_resource: dict
        The CFN resource that matches the sam metadata resource name
    sam_metadata_resource: Dict
       The sam metadata resource properties
    sam_metadata_resource_address: str
        The sam metadata resource address
    """
    cfn_resource_properties = cfn_resource.get("Properties", {})
    resource_type = sam_metadata_resource.get(SAM_METADATA_RESOURCE_TYPE_ATTRIBUTE)
    cfn_resource_type = cfn_resource.get("Type")
    LOG.debug(
        "Validate if the referenced resource in sam metadata resource %s is of the expected type %s",
        sam_metadata_resource_address,
        resource_type,
    )

    if cfn_resource_type != CFN_AWS_LAMBDA_LAYER_VERSION or not cfn_resource_properties:
        LOG.error(
            "The matched resource is of type %s but the type mentioned in the sam metadata resource %s is %s",
            cfn_resource_type,
            sam_metadata_resource_address,
            resource_type,
        )
        raise InvalidSamMetadataPropertiesException(
            f"The sam metadata resource {sam_metadata_resource_address} is referring to a resource that does not "
            f"match the resource type {resource_type}."
        )


def _get_source_code_path(
    sam_metadata_resource: dict,
    sam_metadata_resource_address: str,
    project_root_dir: str,
    src_code_property_name: str,
    property_path_property_name: str,
    src_code_attribute_name: str,
) -> str:
    """
    Validate that sam metadata resource contains the valid metadata properties
    to get a lambda function or layer source code.

    Parameters
    ----------
    sam_metadata_resource: Dict
        The sam metadata resource properties
    sam_metadata_resource_address: str
        The sam metadata resource address
    project_root_dir: str
        the terraform project root directory path
    src_code_property_name: str
        the sam metadata property name that contains the lambda function or layer source code or docker context path
    property_path_property_name: str
        the sam metadata property name that contains the property to get the source code value if it was provided
        as json string
    src_code_attribute_name: str
        the lambda function or later source code or docker context to be used to raise the correct exception

    Returns
    -------
    str
        The lambda function or layer source code or docker context paths
    """
    LOG.debug(
        "Extract the %s from the sam metadata resource %s from property %s",
        src_code_attribute_name,
        sam_metadata_resource_address,
        src_code_property_name,
    )
    source_code = get_sam_metadata_planned_resource_value_attribute(sam_metadata_resource, src_code_property_name)
    source_code_property = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource, property_path_property_name
    )
    LOG.debug(
        "The found %s value is %s and property value is %s", src_code_attribute_name, source_code, source_code_property
    )
    if not source_code:
        raise InvalidSamMetadataPropertiesException(
            f"The sam metadata resource {sam_metadata_resource_address} "
            f"should contain the lambda function/lambda layer "
            f"{src_code_attribute_name} in property {src_code_property_name}"
        )
    if isinstance(source_code, str):
        try:
            LOG.debug("Try to decode the %s value in case if it is a encoded JSON string.", src_code_attribute_name)
            source_code = json.loads(source_code)
            LOG.debug("The decoded value of the %s value is %s", src_code_attribute_name, source_code)
        except JSONDecodeError:
            LOG.debug("Source code value could not be parsed as a JSON object. Handle it as normal string value")
            cfn_source_code_path = source_code

    if isinstance(source_code, list):
        # SAM CLI does not process multiple paths, so we will handle only the first value in this list
        # The first value can either be a string or dict so update source_code to be the first element of the list
        LOG.debug(
            "Process the extracted %s as list, and get the first value as SAM CLI does not support multiple paths",
            src_code_attribute_name,
        )
        if len(source_code) < 1:
            raise InvalidSamMetadataPropertiesException(
                f"The sam metadata resource {sam_metadata_resource_address} "
                f"should contain the lambda function/lambda layer "
                f"{src_code_attribute_name} in property {src_code_property_name}, and it should not be an empty list"
            )
        source_code = source_code[0]
        if not source_code:
            raise InvalidSamMetadataPropertiesException(
                f"The sam metadata resource {sam_metadata_resource_address} "
                f"should contain a valid lambda/lambda layer function "
                f"{src_code_attribute_name} in property {src_code_property_name}"
            )
    if isinstance(source_code, dict):
        LOG.debug(
            "Process the extracted %s as JSON object using the property %s",
            src_code_attribute_name,
            source_code_property,
        )
        if not source_code_property:
            raise InvalidSamMetadataPropertiesException(
                f"The sam metadata resource {sam_metadata_resource_address} "
                f"should contain the lambda function/lambda layer "
                f"{src_code_attribute_name} property in property {property_path_property_name} as the "
                f"{src_code_property_name} value is an object"
            )
        cfn_source_code_path = source_code.get(source_code_property)
        if not cfn_source_code_path:
            LOG.error(
                "The property %s does not exist in the extracted %s JSON object %s",
                source_code_property,
                src_code_attribute_name,
                source_code,
            )
            raise InvalidSamMetadataPropertiesException(
                f"The sam metadata resource {sam_metadata_resource_address} "
                f"should contain a valid lambda function/lambda layer "
                f"{src_code_attribute_name} property in property {property_path_property_name} as the "
                f"{src_code_property_name} value is an object"
            )
    else:
        cfn_source_code_path = source_code

    LOG.debug("The %s path value is %s", src_code_attribute_name, cfn_source_code_path)

    if not os.path.isabs(cfn_source_code_path):
        LOG.debug(
            "The %s path value is not absoulte value. Get the absolute value based on the root directory %s",
            src_code_attribute_name,
            project_root_dir,
        )
        cfn_source_code_path = os.path.normpath(os.path.join(project_root_dir, cfn_source_code_path))
        LOG.debug("The calculated absolute path of %s is %s", src_code_attribute_name, cfn_source_code_path)

    if not isinstance(cfn_source_code_path, str) or not os.path.exists(cfn_source_code_path):
        LOG.error("The path %s does not exist", cfn_source_code_path)
        raise InvalidSamMetadataPropertiesException(
            f"The sam metadata resource {sam_metadata_resource_address} should contain a valid string value for the "
            f"lambda function/lambda layer {src_code_attribute_name} path"
        )

    return cfn_source_code_path


def _get_relevant_cfn_resource(
    sam_metadata_resource: SamMetadataResource,
    cfn_resources: Dict[str, Dict],
    lambda_resources_to_code_map: Dict[str, List[Tuple[Dict, str]]],
) -> List[Tuple[Dict, str]]:
    """
    use the sam metadata resource name property to determine the resource address, and transform the address to logical
    id to use it to get the cfn_resource.
    If the metadata resource does not contain a resource name property, so we need to use the resource built artifact
    path to find tha lambda resources that use the same artifact path

    Parameters
    ----------
    sam_metadata_resource: SamMetadataResource
        sam metadata resource that contain extra information about some resource.
    cfn_resources: Dict
        CloudFormation resources
    lambda_resources_to_code_map: Dict
        The map between lambda resources code path, and lambda resources logical ids

    Returns
    -------
    List[tuple(Dict, str)]
        The cfn resources that mentioned in the sam metadata resource, and the resource logical id
    """

    resources_types = {
        "ZIP_LAMBDA_FUNCTION": "zip",
        "IMAGE_LAMBDA_FUNCTION": "image",
        "LAMBDA_LAYER": "layer",
    }

    sam_metadata_resource_address = sam_metadata_resource.resource.get("address")
    resource_name = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource.resource, SAM_METADATA_RESOURCE_NAME_ATTRIBUTE
    )
    resource_type = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource.resource, SAM_METADATA_RESOURCE_TYPE_ATTRIBUTE
    )
    if not resource_name:
        artifact_property_name = (
            "built_output_path" if resource_type in ["ZIP_LAMBDA_FUNCTION", "LAMBDA_LAYER"] else "built_image_uri"
        )
        artifact_path_value = get_sam_metadata_planned_resource_value_attribute(
            sam_metadata_resource.resource, artifact_property_name
        )
        if not artifact_path_value:
            artifact_path_value = _resolve_resource_attribute(
                sam_metadata_resource.config_resource, artifact_property_name
            )
        hash_value = (
            f"{resources_types[resource_type]}_{_calculate_configuration_attribute_value_hash(artifact_path_value)}"
        )
        lambda_resources = lambda_resources_to_code_map.get(hash_value, [])
        if not lambda_resources:
            raise InvalidSamMetadataPropertiesException(
                f"sam cli expects the sam metadata resource {sam_metadata_resource_address} to contain a resource name "
                f"that will be enriched using this metadata resource"
            )
        return lambda_resources
    # the provided resource name will be always a postfix to the module address. The customer could not set a full
    # address within a module.
    LOG.debug(
        "Check if the input resource name %s is a postfix to the current module address %s",
        resource_name,
        sam_metadata_resource.current_module_address,
    )
    full_resource_address = (
        f"{sam_metadata_resource.current_module_address}.{resource_name}"
        if sam_metadata_resource.current_module_address
        else resource_name
    )
    LOG.debug("check if the resource address %s has a relevant cfn resource or not", full_resource_address)
    logical_id = build_cfn_logical_id(full_resource_address)
    cfn_resource = cfn_resources.get(logical_id)
    if cfn_resource:
        LOG.debug("The CFN resource that match the input resource name %s is %s", resource_name, logical_id)
        return [(cfn_resource, logical_id)]

    raise InvalidSamMetadataPropertiesException(
        f"There is no resource found that match the provided resource name " f"{resource_name}"
    )


def _set_zip_metadata_resources(
    resource: dict,
    cfn_source_code_path: str,
    output_directory_path: str,
    terraform_application_dir: str,
    code_property: str,
) -> None:
    """
    Update the CloudFormation resource metadata with the enrichment properties from the TF resource

    Parameters
    ----------
    resource: dict
        The CFN resource that matches the sam metadata resource name
    cfn_source_code_path: dict
        Absolute path location of where the original source code resides.
    output_directory_path: str
        The directory where to find the Makefile the path to be copied into the temp dir.
    terraform_application_dir: str
        The working directory from which to run the Makefile.
    code_property:
        The property in the configuration used to denote the code e.g. "Code" or "Content"
    """
    resource_properties = resource.get("Properties", {})
    resource_properties[code_property] = cfn_source_code_path
    if not resource.get("Metadata", {}):
        resource["Metadata"] = {}
    resource["Metadata"]["SkipBuild"] = False
    resource["Metadata"]["BuildMethod"] = "makefile"
    resource["Metadata"]["ContextPath"] = output_directory_path
    resource["Metadata"]["WorkingDirectory"] = terraform_application_dir
    # currently we set the terraform project root directory that contains all the terraform artifacts as the project
    # directory till we work on the custom hook properties, and add a property for this value.
    resource["Metadata"]["ProjectRootDirectory"] = terraform_application_dir


def _validate_referenced_resource_matches_sam_metadata_type(
    cfn_resource: dict, sam_metadata_resource: dict, sam_metadata_resource_address: str, expected_package_type: str
) -> None:
    """
    Validate if the resource that match the resource name provided in the sam metadata resource matches the resource
    type provided in the metadata as well.

    Parameters
    ----------
    cfn_resource: dict
        The CFN resource that matches the sam metadata resource name
    sam_metadata_resource: Dict
        The sam metadata resource properties
    sam_metadata_resource_address: str
        The sam metadata resource address
    expected_package_type: str
        The expected lambda function package type.
    """
    cfn_resource_properties = cfn_resource.get("Properties", {})
    resource_type = get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource, SAM_METADATA_RESOURCE_TYPE_ATTRIBUTE
    )
    cfn_resource_type = cfn_resource.get("Type")
    lambda_function_package_type = cfn_resource_properties.get("PackageType", ZIP)
    LOG.debug(
        "Validate if the referenced resource in sam metadata resource %s is of the expected type %s",
        sam_metadata_resource_address,
        resource_type,
    )

    if (
        cfn_resource_type != CFN_AWS_LAMBDA_FUNCTION
        or not cfn_resource_properties
        or lambda_function_package_type != expected_package_type
    ):
        LOG.error(
            "The matched resource is of type %s, and package type is %s, but the type mentioned in the sam metadata "
            "resource %s is %s",
            cfn_resource_type,
            lambda_function_package_type,
            sam_metadata_resource_address,
            resource_type,
        )
        raise InvalidSamMetadataPropertiesException(
            f"The sam metadata resource {sam_metadata_resource_address} is referring to a resource that does not "
            f"match the resource type {resource_type}."
        )


def _get_python_command_name() -> str:
    """
    Verify that python is installed and return the name of the python command

    Returns
    -------
    str
        The name of the python command installed
    """
    command_names_to_try = ["python3", "py3", "python", "py"]
    for command_name in command_names_to_try:
        try:
            run_result = run([command_name, "--version"], check=True, capture_output=True, text=True)
        except CalledProcessError:
            pass
        except OSError:
            pass
        else:
            # check python version
            if not PYTHON_VERSION_REGEX.match(run_result.stdout):
                continue
            return command_name
    raise PrepareHookException("Python not found. Please ensure that python 3.7 or above is installed.")



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
YAML helper, sourced from the AWS CLI

https://github.com/aws/aws-cli/blob/develop/awscli/customizations/cloudformation/yamlhelper.py
"""
# pylint: disable=too-many-ancestors

import json
from typing import Dict, Optional, cast

import yaml
from botocore.compat import OrderedDict
from samtranslator.utils.py27hash_fix import Py27Dict, Py27UniStr
from yaml.nodes import ScalarNode, SequenceNode

TAG_STR = "tag:yaml.org,2002:str"
TIMESTAMP_TAG = "tag:yaml.org,2002:timestamp"


def string_representer(dumper, value):
    """
    Customer Yaml representer that will force the scalar to be quoted in a yaml.dump
    if it scalar starts with a 0. This is needed to keep account ids a string instead
    of turning into on int because yaml thinks it an octal.

    Parameters
    ----------
    dumper yaml.dumper
    value str
        Value in template to resolve

    Returns
    -------

    """
    if value.startswith("0"):
        return dumper.represent_scalar(TAG_STR, value, style="'")

    return dumper.represent_scalar(TAG_STR, value)


def intrinsics_multi_constructor(loader, tag_prefix, node):
    """
    YAML constructor to parse CloudFormation intrinsics.
    This will return a dictionary with key being the instrinsic name
    """

    # Get the actual tag name excluding the first exclamation
    tag = node.tag[1:]

    # Some intrinsic functions doesn't support prefix "Fn::"
    prefix = "Fn::"
    if tag in ["Ref", "Condition"]:
        prefix = ""

    cfntag = prefix + tag

    if tag == "GetAtt" and isinstance(node.value, str):
        # ShortHand notation for !GetAtt accepts Resource.Attribute format
        # while the standard notation is to use an array
        # [Resource, Attribute]. Convert shorthand to standard format
        value = node.value.split(".", 1)

    elif isinstance(node, ScalarNode):
        # Value of this node is scalar
        value = loader.construct_scalar(node)

    elif isinstance(node, SequenceNode):
        # Value of this node is an array (Ex: [1,2])
        value = loader.construct_sequence(node)

    else:
        # Value of this node is an mapping (ex: {foo: bar})
        value = loader.construct_mapping(node)

    return {cfntag: value}


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def yaml_dump(dict_to_dump):
    """
    Dumps the dictionary as a YAML document
    :param dict_to_dump:
    :return:
    """
    CfnDumper.add_representer(OrderedDict, _dict_representer)
    CfnDumper.add_representer(str, string_representer)
    CfnDumper.add_representer(Py27Dict, _dict_representer)
    CfnDumper.add_representer(Py27UniStr, string_representer)
    return yaml.dump(dict_to_dump, default_flow_style=False, Dumper=CfnDumper)


def _dict_constructor(loader, node):
    # Necessary in order to make yaml merge tags work
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


def yaml_parse(yamlstr) -> Dict:
    """Parse a yaml string"""
    try:
        # PyYAML doesn't support json as well as it should, so if the input
        # is actually just json it is better to parse it with the standard
        # json parser.
        return cast(Dict, json.loads(yamlstr, object_pairs_hook=OrderedDict))
    except ValueError:
        yaml.constructor.SafeConstructor.yaml_constructors[
            TIMESTAMP_TAG
        ] = yaml.constructor.SafeConstructor.yaml_constructors[TAG_STR]
        yaml.SafeLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, _dict_constructor)
        yaml.SafeLoader.add_multi_constructor("!", intrinsics_multi_constructor)
        return cast(Dict, yaml.safe_load(yamlstr))


def parse_yaml_file(file_path, extra_context: Optional[Dict] = None) -> Dict:
    """
    Read the file, do variable substitution, parse it as JSON/YAML

    Parameters
    ----------
    file_path : string
        Path to the file to read
    extra_context : Dict
        if the file contains variable in the format of %(variableName)s i.e. the same format of the string % operator,
        this parameter provides the values for those variables substitution.

    Returns
    -------
    questions data as a dictionary
    """

    with open(file_path, "r", encoding="utf-8") as fp:
        content = fp.read()
        if isinstance(extra_context, dict):
            content = content % extra_context
        return yaml_parse(content)


class CfnDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True



"""
Default Settings used by the CLI.

We will checkin the development.py file into source control. So by default only the dev configs
will be available. When preparing the CLI for production release, the release process will inject
production.py file into this folder and remove development.py. When customers install SAM CLI from
PyPi or any other official installation mechanism, they will get the production settings.

Ensure the configuration variables defined in production.py and development.py have exact same names.


Following variables are exported by this module:

    ``telemetry_endpoint_url``: string URL where Telemetry data should be published to

"""

import os

telemetry_endpoint_url = os.getenv(
    "__SAM_CLI_TELEMETRY_ENDPOINT_URL", "https://aws-serverless-tools-telemetry.us-west-2.amazonaws.com/metrics"
)



"""
SAM CLI Default Build constants
"""
import os

DEFAULT_STACK_NAME = "sam-app"
DEFAULT_BUILD_DIR = os.path.join(".aws-sam", "build")
DEFAULT_BUILD_DIR_WITH_AUTO_DEPENDENCY_LAYER = os.path.join(".aws-sam", "auto-dependency-layer")
DEFAULT_CACHE_DIR = os.path.join(".aws-sam", "cache")
DEFAULT_BUILT_TEMPLATE_PATH = os.path.join(".aws-sam", "build", "template.yaml")



"""
Common Path related utilities
"""
from pathlib import PureWindowsPath


def convert_path_to_unix_path(path: str) -> str:
    """
    Convert a path to a unix format path if it is windows, and leave it as is if it in a unix format.
    Parameters
    ----------
    path: str
        the path to be converted

    Returns
    -------
    str
        the path in unix format
    """
    return PureWindowsPath(path).as_posix()



"""
CLI command for "delete" command
"""

import logging
from typing import Optional

import click

from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.lib.utils.version_checker import check_newer_version

SHORT_HELP = "Delete an AWS SAM application and the artifacts created by sam deploy."

HELP_TEXT = """The sam delete command deletes the CloudFormation
stack and all the artifacts which were created using sam deploy.

\b
e.g. sam delete

\b
"""

LOG = logging.getLogger(__name__)


@click.command(
    "delete",
    short_help=SHORT_HELP,
    context_settings={"ignore_unknown_options": False, "allow_interspersed_args": True, "allow_extra_args": True},
    help=HELP_TEXT,
)
@click.option(
    "--stack-name",
    required=False,
    help="The name of the AWS CloudFormation stack you want to delete. ",
)
@click.option(
    "--config-file",
    help=(
        "The path and file name of the configuration file containing default parameter values to use. "
        "Its default value is 'samconfig.toml' in project directory. For more information about configuration files, "
        "see: "
        "https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-config.html."
    ),
    type=click.STRING,
    default="samconfig.toml",
    show_default=True,
)
@click.option(
    "--config-env",
    help=(
        "The environment name specifying the default parameter values in the configuration file to use. "
        "Its default value is 'default'. For more information about configuration files, see: "
        "https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-config.html."
    ),
    type=click.STRING,
    default="default",
    show_default=True,
)
@click.option(
    "--no-prompts",
    help=("Specify this flag to allow SAM CLI to skip through the guided prompts."),
    is_flag=True,
    required=False,
)
@click.option(
    "--s3-bucket",
    help=("The S3 bucket path you want to delete."),
    type=click.STRING,
    default=None,
    required=False,
)
@click.option(
    "--s3-prefix",
    help=("The S3 prefix you want to delete"),
    type=click.STRING,
    default=None,
    required=False,
)
@aws_creds_options
@common_options
@pass_context
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(ctx, stack_name: str, config_file: str, config_env: str, no_prompts: bool, s3_bucket: str, s3_prefix: str):
    """
    `sam delete` command entry point
    """

    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing
    do_cli(
        stack_name=stack_name,
        region=ctx.region,
        config_file=config_file,
        config_env=config_env,
        profile=ctx.profile,
        no_prompts=no_prompts,
        s3_bucket=s3_bucket,
        s3_prefix=s3_prefix,
    )  # pragma: no cover


def do_cli(
    stack_name: str,
    region: str,
    config_file: str,
    config_env: str,
    profile: str,
    no_prompts: bool,
    s3_bucket: Optional[str],
    s3_prefix: Optional[str],
):
    """
    Implementation of the ``cli`` method
    """
    from samcli.commands.delete.delete_context import DeleteContext

    with DeleteContext(
        stack_name=stack_name,
        region=region,
        profile=profile,
        config_file=config_file,
        config_env=config_env,
        no_prompts=no_prompts,
        s3_bucket=s3_bucket,
        s3_prefix=s3_prefix,
    ) as delete_context:
        delete_context.run()



"""Module for storing information about existing workflows."""

from collections import namedtuple
from typing import List

CONFIG = namedtuple(
    "Capability",
    ["language", "dependency_manager", "application_framework", "manifest_name", "executable_search_paths"],
)

PYTHON_PIP_CONFIG = CONFIG(
    language="python",
    dependency_manager="pip",
    application_framework=None,
    manifest_name="requirements.txt",
    executable_search_paths=None,
)

NODEJS_NPM_CONFIG = CONFIG(
    language="nodejs",
    dependency_manager="npm",
    application_framework=None,
    manifest_name="package.json",
    executable_search_paths=None,
)

RUBY_BUNDLER_CONFIG = CONFIG(
    language="ruby",
    dependency_manager="bundler",
    application_framework=None,
    manifest_name="Gemfile",
    executable_search_paths=None,
)

JAVA_GRADLE_CONFIG = CONFIG(
    language="java",
    dependency_manager="gradle",
    application_framework=None,
    manifest_name="build.gradle",
    executable_search_paths=None,
)

JAVA_KOTLIN_GRADLE_CONFIG = CONFIG(
    language="java",
    dependency_manager="gradle",
    application_framework=None,
    manifest_name="build.gradle.kts",
    executable_search_paths=None,
)

JAVA_MAVEN_CONFIG = CONFIG(
    language="java",
    dependency_manager="maven",
    application_framework=None,
    manifest_name="pom.xml",
    executable_search_paths=None,
)

DOTNET_CLIPACKAGE_CONFIG = CONFIG(
    language="dotnet",
    dependency_manager="cli-package",
    application_framework=None,
    manifest_name=".csproj",
    executable_search_paths=None,
)

GO_MOD_CONFIG = CONFIG(
    language="go",
    dependency_manager="modules",
    application_framework=None,
    manifest_name="go.mod",
    executable_search_paths=None,
)

PROVIDED_MAKE_CONFIG = CONFIG(
    language="provided",
    dependency_manager=None,
    application_framework=None,
    manifest_name="Makefile",
    executable_search_paths=None,
)

NODEJS_NPM_ESBUILD_CONFIG = CONFIG(
    language="nodejs",
    dependency_manager="npm-esbuild",
    application_framework=None,
    manifest_name="package.json",
    executable_search_paths=None,
)

RUST_CARGO_LAMBDA_CONFIG = CONFIG(
    language="rust",
    dependency_manager="cargo",
    application_framework=None,
    manifest_name="Cargo.toml",
    executable_search_paths=None,
)

ALL_CONFIGS: List[CONFIG] = [
    PYTHON_PIP_CONFIG,
    NODEJS_NPM_CONFIG,
    RUBY_BUNDLER_CONFIG,
    JAVA_GRADLE_CONFIG,
    JAVA_KOTLIN_GRADLE_CONFIG,
    JAVA_MAVEN_CONFIG,
    DOTNET_CLIPACKAGE_CONFIG,
    GO_MOD_CONFIG,
    PROVIDED_MAKE_CONFIG,
    NODEJS_NPM_ESBUILD_CONFIG,
    RUST_CARGO_LAMBDA_CONFIG,
]



"""
Invokable Module for CLI

python -m samcli
"""

from samcli.cli.main import cli  # pragma: no cover

if __name__ == "__main__":  # pragma: no cover
    # NOTE(TheSriram): prog_name is always set to "sam". This way when the CLI is invoked as a module,
    # the help text that is generated still says "sam" instead of "__main__".
    cli(prog_name="sam")



"""A flow of questions to be asked to the user in an interactive way."""
from typing import Any, Dict, List, Optional, Tuple

import click

from ..utils.colors import Colored
from .question import Question


class InteractiveFlow:
    """
    This class depends on the click module to prompt the user for answers to the questions of the flow.

    Attributes
    ----------
    _questions: Dict[str, Question]
        The questions to prompt the user for its answers
    _first_question_key: str
        The entry-point of the flow, it tells which question to ask the user first
    """

    def __init__(self, questions: Dict[str, Question], first_question_key: str):
        self._questions: Dict[str, Question] = questions
        self._first_question_key: str = first_question_key
        self._current_question: Optional[Question] = None
        self._color = Colored()

    def advance_to_next_question(self, current_answer: Optional[Any] = None) -> Optional[Question]:
        """
        Based on the answer of the current question, what is the next question to ask.

        Parameters
        ----------
        current_answer: str
            User's answer of the current question. This can be None of this is the first question or an optional
            question and the user didn't provide an answer to

        Returns: The next question defined in current question's next-question map or the first question if this
                is the start of the flow.
        """
        if not self._current_question:
            self._current_question = self._questions.get(self._first_question_key)
        else:
            next_question_key = self._current_question.get_next_question_key(current_answer)
            self._current_question = self._questions.get(next_question_key) if next_question_key else None
        return self._current_question

    def run(
        self,
        context: Dict,
    ) -> Dict:
        """
        starts the flow, collects user's answers to the question and return a new copy of the passed context
        with the answers appended to the copy

        Parameters
        ----------
        context: Dict
            The cookiecutter context before prompting this flow's questions
            The context can be used to provide default values, and support both str keys and List[str] keys.

        Returns
        -------
        A new copy of the context with user's answers added to the copy such that each answer is
             associated to the key of the corresponding question
        """
        context = context.copy()
        answers: List[Tuple[str, Any]] = []

        question = self.advance_to_next_question()
        while question:
            answer = question.ask(context=context)
            context[question.key] = answer
            answers.append((question.key, answer))
            question = self.advance_to_next_question(answer)

        # print summary
        click.echo(self._color.bold("SUMMARY"))
        click.echo("We will generate a pipeline config file based on the following information:")

        for question_key, answer in answers:
            if answer is None:
                # ignore unanswered questions
                continue

            question = self._questions[question_key]
            click.echo(f"\t{question.text}: {self._color.underline(str(answer))}")

        return context



import json

# import requests


def lambda_handler(event, context):
    """Sample pure Lambda function

    Parameters
    ----------
    event: dict, required
        API Gateway Lambda Proxy Input Format

        Event doc: https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html#api-gateway-simple-proxy-for-lambda-input-format

    context: object, required
        Lambda Context runtime methods and attributes

        Context doc: https://docs.aws.amazon.com/lambda/latest/dg/python-context-object.html

    Returns
    ------
    API Gateway Lambda Proxy Output Format: dict

        Return doc: https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html
    """

    # try:
    #     ip = requests.get("http://checkip.amazonaws.com/")
    # except requests.RequestException as e:
    #     # Send some context about this error to Lambda Logs
    #     print(e)

    #     raise e

    return {
        "statusCode": 200,
        "body": json.dumps(
            {
                "message": "hello world",
                # "location": ip.text.replace("\n", "")
            }
        ),
    }



"""
Creates and hashes metadata regarding SAM CLI projects.
"""

import hashlib
import re
import subprocess
from os import getcwd
from os.path import basename
from typing import Optional
from urllib.parse import urlparse

from samcli.cli.global_config import GlobalConfig


def get_git_remote_origin_url() -> Optional[str]:
    """
    Retrieve an hashed version of the project's git remote origin url, if it exists.

    Returns
    -------
    str | None
        A SHA256 hexdigest string of the git remote origin url, formatted such that the
        hashed value follows the pattern <hostname>/<owner>/<project_name>.git.
        If telemetry is opted out of by the user, or the `.git` folder is not found
        (the directory is not a git repository), returns None
    """
    if not bool(GlobalConfig().telemetry_enabled):
        return None

    git_url = None
    try:
        runcmd = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"], capture_output=True, shell=True, check=True, text=True
        )
        git_url = _parse_remote_origin_url(str(runcmd.stdout))
    except subprocess.CalledProcessError:
        # Ignoring, None git_url will be handled later
        pass

    return _hash_value(git_url) if git_url else None


def get_project_name() -> Optional[str]:
    """
    Retrieve an hashed version of the project's name, as defined by the .git folder (or directory name if no .git).

    Returns
    -------
    str | None
        A SHA256 hexdigest string of either the name of the project, or the name of the
        current working directory that the command is running in.
        If telemetry is opted out of by the user, returns None
    """
    if not bool(GlobalConfig().telemetry_enabled):
        return None

    project_name = None
    try:
        runcmd = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"], capture_output=True, shell=True, check=True, text=True
        )
        git_url = _parse_remote_origin_url(str(runcmd.stdout))
        if git_url:
            project_name = git_url.split("/")[-1]  # dir is git repo, get project name from URL
    except subprocess.CalledProcessError:
        # Ignoring, None project_name will be handled at the end before returning
        pass

    if not project_name:
        project_name = basename(getcwd().replace("\\", "/"))  # dir is not a git repo, get directory name

    return _hash_value(project_name)


def get_initial_commit_hash() -> Optional[str]:
    """
    Retrieve an hashed version of the project's initial commit hash, if it exists.

    Returns
    -------
    str | None
        A SHA256 hexdigest string of the git project's initial commit hash.
        If telemetry is opted out of by the user, or the `.git` folder is not found
        (the directory is not a git repository), returns None.
    """
    if not bool(GlobalConfig().telemetry_enabled):
        return None

    metadata = None
    try:
        runcmd = subprocess.run(
            ["git", "rev-list", "--max-parents=0", "HEAD"], capture_output=True, shell=True, check=True, text=True
        )
        metadata = runcmd.stdout.strip()
    except subprocess.CalledProcessError:
        return None  # Not a git repo

    return _hash_value(metadata)


def _parse_remote_origin_url(url: str) -> Optional[str]:
    """
    Parse a `git remote origin url` into a formatted "hostname/project" string

    Returns
    -------
    str
        formatted project origin url
    """
    parsed = urlparse(url)
    if not parsed.path:
        return None

    formatted = (parsed.hostname or "") + parsed.path
    formatted = re.sub(r"\n", "", formatted)
    formatted = re.sub("/$", "", formatted)
    formatted = re.sub(".git$", "", formatted)
    formatted = re.sub("^(.+)@", "", formatted)
    formatted = formatted.replace(":", "/")

    return formatted


def _hash_value(value: str) -> str:
    """Hash a string, and then return the hashed value as a byte string."""
    h = hashlib.sha256()
    h.update(value.encode("utf-8"))
    return h.hexdigest()



"""
Command group for "pipeline" suite commands. It provides common CLI arguments, template parsing capabilities,
setting up stdin/stdout etc
"""

import click

from .bootstrap.cli import cli as bootstrap_cli
from .init.cli import cli as init_cli


@click.group()
def cli() -> None:
    """
    Manage the continuous delivery of the application
    """


# Add individual commands under this group
cli.add_command(bootstrap_cli)
cli.add_command(init_cli)



"""
Display of the Endpoints of a SAM stack
"""
import logging
from typing import Optional

from samcli.commands.list.cli_common.list_common_context import ListContext
from samcli.lib.list.endpoints.endpoints_producer import EndpointsProducer
from samcli.lib.list.list_interfaces import ProducersEnum
from samcli.lib.list.mapper_consumer_factory import MapperConsumerFactory

LOG = logging.getLogger(__name__)


class EndpointsContext(ListContext):
    """
    Context class for sam list endpoints
    """

    def __init__(
        self, stack_name: str, output: str, region: Optional[str], profile: Optional[str], template_file: Optional[str]
    ):
        """
        Parameters
        ----------
        stack_name: str
            The name of the stack
        output: str
            The format of the output, either json or table
        region: Optional[str]
            The region of the stack
        profile: Optional[str]
            Optional profile to be used
        template_file: Optional[str]
            The location of the template file. If one is not specified, the default will be "template.yaml" in the CWD
        """
        super().__init__()
        self.stack_name = stack_name
        self.output = output
        self.region = region
        self.profile = profile
        self.template_file = template_file
        self.iam_client = None
        self.cloudcontrol_client = None
        self.apigateway_client = None
        self.apigatewayv2_client = None

    def __enter__(self):
        self.init_clients()
        return self

    def __exit__(self, *args):
        pass

    def init_clients(self) -> None:
        """
        Initialize the clients being used by sam list.
        """
        super().init_clients()
        self.iam_client = self.client_provider("iam")
        self.cloudcontrol_client = self.client_provider("cloudcontrol")
        self.apigateway_client = self.client_provider("apigateway")
        self.apigatewayv2_client = self.client_provider("apigatewayv2")

    def run(self) -> None:
        """
        Get the resources for a stack
        """
        factory = MapperConsumerFactory()
        container = factory.create(producer=ProducersEnum.ENDPOINTS_PRODUCER, output=self.output)
        endpoints_producer = EndpointsProducer(
            stack_name=self.stack_name,
            region=self.region,
            profile=self.profile,
            template_file=self.template_file,
            cloudformation_client=self.cloudformation_client,
            iam_client=self.iam_client,
            cloudcontrol_client=self.cloudcontrol_client,
            apigateway_client=self.apigateway_client,
            apigatewayv2_client=self.apigatewayv2_client,
            mapper=container.mapper,
            consumer=container.consumer,
        )
        endpoints_producer.produce()



"""
Sets up the cli for stack-outputs
"""

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands.list.cli_common.options import output_option
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

HELP_TEXT = """
Get the stack outputs as defined in the SAM/CloudFormation template.
"""


@click.command(name="stack-outputs", help=HELP_TEXT)
@click.option(
    "--stack-name",
    help="Name of corresponding deployed stack. ",
    required=True,
    type=click.STRING,
)
@configuration_option(provider=TomlProvider(section="parameters"))
@output_option
@aws_creds_options
@common_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(self, stack_name, output, config_file, config_env):
    """
    `sam list stack-outputs` command entry point
    """
    do_cli(stack_name=stack_name, output=output, region=self.region, profile=self.profile)


def do_cli(stack_name, output, region, profile):
    """
    Implementation of the ``cli`` method
    """
    from samcli.commands.list.stack_outputs.stack_outputs_context import StackOutputsContext

    with StackOutputsContext(
        stack_name=stack_name, output=output, region=region, profile=profile
    ) as stack_output_context:
        stack_output_context.run()



"""
Provide a factory class for IaC project creation
"""
import fnmatch
import os

from samcli.lib.iac.cdk.cdk_iac import CdkIacImplementation
from samcli.lib.iac.cfn.cfn_iac import CfnIacImplementation
from samcli.lib.iac.exceptions import InvalidIaCPluginException, InvalidProjectTypeException
from samcli.lib.iac.plugins_interfaces import IaCPluginInterface, ProjectTypes, SamCliContext

IAC_IMPLEMENTATIONS = {
    ProjectTypes.CDK.value: CdkIacImplementation,
    ProjectTypes.CFN.value: CfnIacImplementation,
}


class IaCFactory:
    """
    Generate the appropriate IaC implementation
    get_iac returns an instance of an IaCPluginInterface based on the given context
    detect_project_type returns a string indicating the type of IaC project found
    """

    def __init__(self, context: SamCliContext):
        self._sam_cli_context = context

    def get_iac(self) -> IaCPluginInterface:
        project_type_string = "project_type"
        if project_type_string not in self._sam_cli_context.command_options_map:
            raise ValueError("Project type not found in sam-cli command options")
        project_type = self._sam_cli_context.command_options_map.get(project_type_string)
        if project_type not in IAC_IMPLEMENTATIONS:
            raise InvalidProjectTypeException(
                msg=f"{project_type} is an invalid project type option value, the value should be one "
                f"of the following {[ptype.value for ptype in ProjectTypes]} ",
            )
        iac_implementation = IAC_IMPLEMENTATIONS.get(project_type)
        if iac_implementation is None:
            raise ValueError("IaC implementation type not found in list of valid IaC implementations.")
        return iac_implementation(self._sam_cli_context)

    @staticmethod
    def detect_project_type(path: str) -> str:
        curr_files = os.listdir(path)
        matched_types = []
        for project_type, implementation in IAC_IMPLEMENTATIONS.items():
            iac_file_patterns = implementation.get_iac_file_patterns()
            # Check if all the files for the IaC type are in the current directory
            for pattern in iac_file_patterns:
                if fnmatch.filter(curr_files, pattern):
                    matched_types.append(project_type)
                    break
        if len(matched_types) != 1:
            raise InvalidIaCPluginException(curr_files)
        return matched_types[0]



"""
Interface for MapperConsumerFactory, Producer, Mapper, ListInfoPullerConsumer
"""
from abc import ABC, abstractmethod
from enum import Enum
from typing import Generic, TypeVar

InputType = TypeVar("InputType")
OutputType = TypeVar("OutputType")


class ListInfoPullerConsumer(ABC, Generic[InputType]):
    """
    Interface definition to consume and display data
    """

    @abstractmethod
    def consume(self, data: InputType):
        """
        Parameters
        ----------
        data: TypeVar
            Data for the consumer to print
        """


class Mapper(ABC, Generic[InputType, OutputType]):
    """
    Interface definition to map data to json or table
    """

    @abstractmethod
    def map(self, data: InputType) -> OutputType:
        """
        Parameters
        ----------
        data: TypeVar
            Data for the mapper to map

        Returns
        -------
        Any
            Mapped output given the data
        """


class Producer(ABC):
    """
    Interface definition to produce data for the mappers and consumers
    """

    mapper: Mapper
    consumer: ListInfoPullerConsumer

    @abstractmethod
    def produce(self):
        """
        Produces the data for the mappers and consumers
        """


class MapperConsumerFactoryInterface(ABC):
    """
    Interface definition to create mapper-consumer factories
    """

    @abstractmethod
    def create(self, producer, output):
        """
        Parameters
        ----------
        producer: str
            A string indicating which producer is calling the function
        output: str
            A string indicating the output type

        Returns
        -------
        MapperConsumerContainer
            A container that contains a mapper and a consumer
        """


class ProducersEnum(Enum):
    STACK_OUTPUTS_PRODUCER = 1
    RESOURCES_PRODUCER = 2
    ENDPOINTS_PRODUCER = 3



"""
CLI command for "local start-api" command
"""

import logging

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, pass_context, print_cmdline_args
from samcli.cli.main import common_options as cli_framework_options
from samcli.commands._utils.option_value_processor import process_image_options
from samcli.commands._utils.options import generate_next_command_recommendation
from samcli.commands.local.cli_common.options import (
    invoke_common_options,
    local_common_options,
    service_common_options,
    warm_containers_common_options,
)
from samcli.commands.local.lib.exceptions import InvalidIntermediateImageError
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version
from samcli.local.docker.exceptions import ContainerNotStartableException

LOG = logging.getLogger(__name__)

HELP_TEXT = """
Allows you to run your Serverless application locally for quick development & testing.
 When run in a directory that contains your Serverless functions and your AWS SAM template, it will create a local HTTP
 server hosting all of your functions. When accessed (via browser, cli etc), it will launch a Docker container locally
 to invoke the function. It will read the CodeUri property of AWS::Serverless::Function resource to find the path in
 your file system containing the Lambda Function code. This could be the project's root directory for interpreted
 languages like Node & Python, or a build directory that stores your compiled artifacts or a JAR file. If you are using
 a interpreted language, local changes will be available immediately in Docker container on every invoke. For more
 compiled languages or projects requiring complex packing support, we recommended you run your own building solution
and point SAM to the directory or file containing build artifacts.
"""


@click.command(
    "start-api",
    help=HELP_TEXT,
    short_help="Sets up a local endpoint you can use to test your API. Supports hot-reloading "
    "so you don't need to restart this service when you make changes to your function.",
)
@configuration_option(provider=TomlProvider(section="parameters"))
@service_common_options(3000)
@click.option(
    "--static-dir",
    "-s",
    default="public",
    help="Any static assets (e.g. CSS/Javascript/HTML) files located in this directory " "will be presented at /",
)
@invoke_common_options
@warm_containers_common_options
@local_common_options
@cli_framework_options
@aws_creds_options  # pylint: disable=R0914
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
def cli(
    ctx,
    # start-api Specific Options
    host,
    port,
    static_dir,
    # Common Options for Lambda Invoke
    template_file,
    env_vars,
    debug_port,
    debug_args,
    debugger_path,
    container_env_vars,
    docker_volume_basedir,
    docker_network,
    log_file,
    layer_cache_basedir,
    skip_pull_image,
    force_image_build,
    parameter_overrides,
    config_file,
    config_env,
    warm_containers,
    shutdown,
    debug_function,
    container_host,
    container_host_interface,
    invoke_image,
):
    """
    `sam local start-api` command entry point
    """
    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing

    do_cli(
        ctx,
        host,
        port,
        static_dir,
        template_file,
        env_vars,
        debug_port,
        debug_args,
        debugger_path,
        container_env_vars,
        docker_volume_basedir,
        docker_network,
        log_file,
        layer_cache_basedir,
        skip_pull_image,
        force_image_build,
        parameter_overrides,
        warm_containers,
        shutdown,
        debug_function,
        container_host,
        container_host_interface,
        invoke_image,
    )  # pragma: no cover


def do_cli(  # pylint: disable=R0914
    ctx,
    host,
    port,
    static_dir,
    template,
    env_vars,
    debug_port,
    debug_args,
    debugger_path,
    container_env_vars,
    docker_volume_basedir,
    docker_network,
    log_file,
    layer_cache_basedir,
    skip_pull_image,
    force_image_build,
    parameter_overrides,
    warm_containers,
    shutdown,
    debug_function,
    container_host,
    container_host_interface,
    invoke_image,
):
    """
    Implementation of the ``cli`` method, just separated out for unit testing purposes
    """

    from samcli.commands.exceptions import UserException
    from samcli.commands.local.cli_common.invoke_context import InvokeContext
    from samcli.commands.local.lib.exceptions import NoApisDefined, OverridesNotWellDefinedError
    from samcli.commands.local.lib.local_api_service import LocalApiService
    from samcli.commands.validate.lib.exceptions import InvalidSamDocumentException
    from samcli.lib.providers.exceptions import InvalidLayerReference
    from samcli.local.docker.lambda_debug_settings import DebuggingNotSupported

    LOG.debug("local start-api command is called")

    processed_invoke_images = process_image_options(invoke_image)

    # Pass all inputs to setup necessary context to invoke function locally.
    # Handler exception raised by the processor for invalid args and print errors

    try:
        with InvokeContext(
            template_file=template,
            function_identifier=None,  # Don't scope to one particular function
            env_vars_file=env_vars,
            docker_volume_basedir=docker_volume_basedir,
            docker_network=docker_network,
            log_file=log_file,
            skip_pull_image=skip_pull_image,
            debug_ports=debug_port,
            debug_args=debug_args,
            debugger_path=debugger_path,
            container_env_vars_file=container_env_vars,
            parameter_overrides=parameter_overrides,
            layer_cache_basedir=layer_cache_basedir,
            force_image_build=force_image_build,
            aws_region=ctx.region,
            aws_profile=ctx.profile,
            warm_container_initialization_mode=warm_containers,
            debug_function=debug_function,
            shutdown=shutdown,
            container_host=container_host,
            container_host_interface=container_host_interface,
            invoke_images=processed_invoke_images,
        ) as invoke_context:

            service = LocalApiService(lambda_invoke_context=invoke_context, port=port, host=host, static_dir=static_dir)
            service.start()
            command_suggestions = generate_next_command_recommendation(
                [
                    ("Validate SAM template", "sam validate"),
                    ("Test Function in the Cloud", "sam sync --stack-name {{stack-name}} --watch"),
                    ("Deploy", "sam deploy --guided"),
                ]
            )
            click.secho(command_suggestions, fg="yellow")

    except NoApisDefined as ex:
        raise UserException(
            "Template does not have any APIs connected to Lambda functions", wrapped_from=ex.__class__.__name__
        ) from ex
    except (
        InvalidSamDocumentException,
        OverridesNotWellDefinedError,
        InvalidLayerReference,
        InvalidIntermediateImageError,
        DebuggingNotSupported,
    ) as ex:
        raise UserException(str(ex), wrapped_from=ex.__class__.__name__) from ex
    except ContainerNotStartableException as ex:
        raise UserException(str(ex), wrapped_from=ex.__class__.__name__) from ex



"""
Exceptions that are raised by sam deploy
"""
from samcli.commands.exceptions import UserException


class ChangeEmptyError(UserException):
    def __init__(self, stack_name):
        self.stack_name = stack_name
        message_fmt = "No changes to deploy. Stack {stack_name} is up to date"
        super().__init__(message=message_fmt.format(stack_name=self.stack_name))


class ChangeSetError(UserException):
    def __init__(self, stack_name, msg):
        self.stack_name = stack_name
        self.msg = msg
        message_fmt = "Failed to create changeset for the stack: {stack_name}, {msg}"
        super().__init__(message=message_fmt.format(stack_name=self.stack_name, msg=self.msg))


class DeployFailedError(UserException):
    def __init__(self, stack_name, msg):
        self.stack_name = stack_name
        self.msg = msg

        message_fmt = "Failed to create/update the stack: {stack_name}, {msg}"

        super().__init__(message=message_fmt.format(stack_name=self.stack_name, msg=msg))


class GuidedDeployFailedError(UserException):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(message=msg)


class DeployStackOutPutFailedError(UserException):
    def __init__(self, stack_name, msg):
        self.stack_name = stack_name
        self.msg = msg

        message_fmt = "Failed to get outputs from stack: {stack_name}, {msg}"

        super().__init__(message=message_fmt.format(stack_name=self.stack_name, msg=msg))


class DeployBucketInDifferentRegionError(UserException):
    def __init__(self, msg):
        self.msg = msg

        message_fmt = "{msg} : deployment s3 bucket is in a different region, try sam deploy --guided"

        super().__init__(message=message_fmt.format(msg=self.msg))


class DeployBucketRequiredError(UserException):
    def __init__(self):

        message_fmt = (
            "Templates with a size greater than 51,200 bytes must be deployed "
            "via an S3 Bucket. Please add the --s3-bucket parameter to your "
            "command. The local template will be copied to that S3 bucket and "
            "then deployed."
        )

        super().__init__(message=message_fmt)


class DeployResolveS3AndS3SetError(UserException):
    def __init__(self):
        message_fmt = (
            "Cannot use both --resolve-s3 and --s3-bucket parameters in non-guided deployments."
            " Please use only one or use the --guided option for a guided deployment."
        )

        super().__init__(message=message_fmt)


class DeployStackStatusMissingError(UserException):
    def __init__(self, stack_name):
        message_fmt = "Was not able to find a stack with the name: {msg}, please check your parameters and try again."
        super().__init__(message=message_fmt.format(msg=stack_name))
